# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************  ModulatorySignal *******************************************************

"""
..
Sections
--------
  * `ModulatorySignal_Overview`
    - `Naming Conventions for ModulatorySignals <ModulatorySignal_Naming>`
  * `ModulatorySignal_Creation`
  * `ModulatorySignal_Structure`
    - `ModulatorySignal_Projections`
    - `ModulatorySignal_Modulation`
      • `ModulatorySignal_Uses`
      • `ModulatorySignal_Types`
      • `ModulatorySignal_Implementation`
  * `ModulatorySignal_Execution`
  * `ModulatorySignal_Class_Reference`
  

.. _ModulatorySignal_Overview:

Overview
--------

A ModulatorySignal is a subclas of `OutputState` that belongs to an `AdaptiveMechanism <AdaptiveMechanism>`, and is
used to `modulate <ModulatorySignal_Modulation>` the `value <State_Base.value>` of one or more `States <State>` by way
of one or more `ModulatoryProjctions <ModulatoryProjection>`. A ModulatorySignal modulates the value of a State by
modifying a  parameter of thatState's `function <State_Base.function>`.  There are three types of ModulatorySignals,
each of which is  associated wth a particular type of `AdaptiveMechanism <AdaptiveMechanism>` and `ModulatoryProjection
<ModulatoryProjection>`, and modifies the value of different types of States, as summarized `below:

* `ControlSignal`
    takes the `allocation <ControlSignal.allocation>` assigned to it by the `function <ControlMechanism.function>`
    of the `ControlMechanism <ControlMechanism>` to which it belongs, and uses it to modulate the parameter of a
    `Mechanism <Mechanism>` or its `function <Mechanism_Base.function>` (and thereby the `value
    <Mechanism_Base.value>` of that Mechanism), or a parameter of the `function <State_Base.function>` one of the
    Mechanism's `InputStates <InputState>` or `OutputStates <OutputState>` (and thereby the `value <State_Base.value>`
    of the corresponding State).
..
* `GatingSignal` takes the `allocation <GatingSignal.allocation>` assigned to it by the `function
    <GatingMechanism.function>` of the `GatingMechanism` to which it belongs, and uses it to modulate the parameter
    of the `function <State_Base.function>` of an `InputState` or `OutputState` (and hence that State's `value
    <State_Base.value>`).  A GatingMechanism and GatingSignal can be thought of as implementing a form of control
    specialized for gating the input to and/or output of a Mechanism.
..
* `LearningSignal`
    takes the `learning_signal <LearningMechanism>` calculated by the `function <LearningMechanism.function>` of the
    `LearningMechanism` to which it belongs, and uses it to modulate the `matrix <MappingProjection.matrix>` parameter
    of a `MappingProjection`.

These are shown in a `figure <ModulatorySignal_Anatomy_Figure>` below, and are descdribed in greater detail in
in the sections under `ModulatorySignal_Structure`.

.. _ModulatorySignal_Naming:

*Naming Conventions for ModulatorySignals*

Modulatory Components and their attributes are named according to the category of modulation:

    - AdaptiveMechanism name: <*Category*>Mechanism (e.g., ControlMechanism)
    - ModulatorySignal name: <*Category*>Signal (e.g., ControlSignal)
    - ModulatoryProjection name: <*Category*>Projection (e.g., ControlProjection)
    - List of an AdaptiveMechanism's ModulatorySignals: <*CategoryMechanism*>.<category>_signals
      (e.g., ControlMechanism.control_signals)
    - Value of a ModulatorySignal: <*CategorySignal*>.<category>_signal (e.g., ControlSignal.control_signal)

.. _ModulatorySignal_Creation:

Creating a ModulatorySignal
---------------------------

ModulatorySignal is a base class, and cannot be instantiated directly.  However, the three types of ModulatorySignals
listed above can be created directly, by calling the constructor for the desired type.  More commonly, however,
ModulatorySignals are created automatically by the `AdaptiveMechanism <AdaptiveMechanism>` to which they belong, or by
specifying them in the constructor for an `AdaptiveMechanism <AdaptiveMechanism>` (the details of which are described in
the documentation for each type of ModulatorySignal).  If a ModulatorySignal is constructed explicitly, the type of
modulation it uses is specifed in the **modulation** argument of its constructor, using a 2-item tuple that contains
the State to be modulated as the first item, and either the name of the parameter of the State's `function
<State_Base.function>` to be modulated, or a keyword specifying the type of modulation, as the second item (see
`ModulatorySignal_Types` for additional details).

.. _ModulatorySignal_Structure:

Structure
---------

A ModulatorySignal is always assigned to an `AdaptiveMechanism <AdaptiveMechanism>`, and must be assigned to an
AdaptiveMechanism of the appropriate type (`see types of AdaptiveMechanism <AdaptiveMechanism_Types>`).  The
ModulatorySignal receives a `modulatory_allocation` from the AdaptiveMechanism to which it is assigned, that it uses
as the `variable <Function_Base.variable>` for its `function <ModulatorySignal.function>`, the result of which is the
modulatory `value <ModulatorySignal.value>` of the ModulatorySignal.  A ModulatorySignal is associated with one or more
`ModulatoryProjections <ModulatoryProjection>` of the corresponding type, that that receive the ModulatorySignal's
`value <ModulatorySignal.value>`, and use this to modulate the State(s) to which they project.  All of the
ModulatoryProjections from a given ModulatorySignal are assigned the same modulatory `value <ModulatorySignal.value>`
(see `ModulatorySignal_Projections` below) and use the same `type of modulation <ModulatorySignal_Types>` specified
by the the ModulatorySignal's `modulation <ModulatorySignal.modulation>` attribute.  The ModulatoryProjections
received by a `State <State>` are listed in the State's `mod_afferents <State_Base.mod_afferents>` attribute.

The section on `Modulation` below provides a comparison of ModulatorySignal `subclasses and their uses
<ModulatorySignal_Uses>` (summarized in an accompanying `table <ModulatorySignal_Table>` and `figure
<ModulatorySignal_Anatomy_Figure>`), as well as a description of the different `types of modulation
<ModulatorySignal_Types>` and a more detailed description of `how modulation is implemented
<ModulatorySignal_Implementation>`.

.. _ModulatorySignal_Projections:

*ModulatoryProjections*
~~~~~~~~~~~~~~~~~~~~~~~

A ModulatorySignal can be assigned one or more `ModulatoryProjections <ModulatoryProjection>`,
using either the **projections** argument of its constructor, or in an entry of a dictionary assigned to the
**params** argument with the key *PROJECTIONS*.  These are assigned to its `efferents  <ModulatorySignal.efferents>`
attribute.  See `State Projections <State_Projections>` for additional details concerning the specification of
Projections when creating a State.

Although a ModulatorySignal can be assigned more than one `ModulatoryProjection <ModulatoryProjection>`,
all of those Projections receive and convey the same modulatory `value <ModulatorySignal.value>` from the
ModulatorySignal, and use the same form of `modulation <ModulatorySignal_Modulation>`.  This is a common use for some
ModulatorySignals (e.g., the use of a single `GatingSignal` to gate multiple `InputState(s) <InputState>` or
`OutputState(s) <OutputState>`), but requires more specialized circumstances for others (e.g., the use of a single
`LearningSignal` for more than one `MappingProjection`, or a single `ControlSignal` for the parameters of more than
one Mechanism or function).

.. _ModulatorySignal_Modulation:

*Modulation*
~~~~~~~~~~~~

A ModulatorySignal modulates the value of a `State <State>` either by modifying a parameter of the State's `function
<State_Base.function>` (which determines the State's `value <State_Base.value>`), or by  assigning a value to the State
directly.  The `type of modulation <ModulatorySignal_Types>` is determined by the ModulatorySignal's
`modulation <ModulatorySignal.modulation>` attribute, which can be specified in the **modulation** argument of its
ModulatorySignal's constructor, or in a *MODULATION* entry of a `State specification dictionary
<State_Specification>` used to create the ModulatorySignal (see `Type of Modualtion <ModulatorySignal_Types>` and
`figure <ModulatorySignal_Detail_Figure>` below for details). If the type of `modulation <ModulatorySignal.modulation>`
is not specified when a ModulatorySignal is created, it is assigned the value of the `modulation
<AdaptiveMechanism_Base.modulation>` attribute for the `AdaptiveMechanism <AdaptiveMechanism>` to which it belongs.

.. _ModulatorySignal_Uses:

Uses of Modulation
^^^^^^^^^^^^^^^^^^

There are three broad categories of modulation that serve different purposes, and differ according to the
ModulatorySignals used and the type of State modulated:

  * **modulation of a** `Mechanism`\\s <function <Mechanism_Base.function> -- a `ControlSignal` must be used; this
    modulates the `ParameterState` for a parameter of the Mechanism's `function <Mechanism_Base.function>` which,
    in turn, determines how it computes the Mechanism's `value <Mechanism_Base.value>`;

  * **modulation of a** `Mechanism`\\s input or output -- a `GatingSignal` is specialized for this purpose, though a
    `ControlSignal` can also be used;  these modulate an `InputState` of the Mechanism, that determines the
    Mechanism's `variable <Mechanism_Base.variable>` used as the input to its `function <Mechanism_Base.function>`,
    or an `OutputState` of the Mechanism, that determines how the `value <Mechanism_Base.value>` of the Mechanism
    (i.e., the result of its `function <Mechanism_Base.function>`) is used to generate the output of the Mechanism.

  * **modulation of a** `MappingProjection` -- a `LearningSignal` must be used; this modulates the `ParameterState` for
    the `matrix <MappingProjection.matrix>` parameter of a MappingProjection's `function  <MappingProjection.function>`
    which, in turn, determines how it computes the MappingProjection's `value <MappingProjection.value>`.

The following table summarizes the three uses of modulation, the ModulatorySignals for each, and the States they
modulate. The mechanics of modulation are described in greater detail in `ModulatorySignal_Implementation`,
and shown in the `figure below <ModulatorySignal_Anatomy_Figure>`.

.. _ModulatorySignal_Table:

.. table:: **ModulatorySignals and States they Modulate**
  :align: left

  +------------------------------------+------------------------+------------------------------+----------------------------------------+----------------------------+
  |                                    |                        |Default type of `modulation   |                                        |Default Function (mod param)|
  |             Purpose                |  ModulatorySignal      |<ModulatorySignal.modulation>`|           Recipient State              |for Recipient State         |
  +====================================+========================+==============================+========================================+============================+
  | Modulate the parameter of a        |                        |                              | Mechanism `ParameterState` (by default)|                            |
  | Mechanism's `function              | `ControlSignal` (blue) |     *MULTIPLICATIVE*         | but can also be an                     |     `Linear` (`slope`)     |
  | <Mechanism_Base.function>`         |                        |                              | `InputState` or `OutputState`          |                            |
  +------------------------------------+------------------------+------------------------------+----------------------------------------+----------------------------+
  | Modulate the input or output of    |                        |                              |                                        |                            |
  | a Mechanism's `function            | `GatingSignal` (brown) |     *MULTIPLICATIVE*         |  Mechanism `InputState`/`OutputState`  |     `Linear` (`slope`)     |
  | <Mechanism_Base.function>`         |                        |                              |                                        |                            |
  +------------------------------------+------------------------+------------------------------+----------------------------------------+----------------------------+
  | Modulate a MappingProjection's     |                        |                              |                                        |   `AccumulatorIntegrator`  |
  | `matrix <MappingProjection.matrix>`|`LearningSignal` (green)|        *ADDITIVE*            |  MappingProjection `ParameterState`    |   (`increment`)            |
  | parameter                          |                        |                              |                                        |                            |
  +------------------------------------+------------------------+------------------------------+----------------------------------------+----------------------------+

Colors listed are those used in the `figure <ModulatorySignal_Anatomy_Figure>` below.

It is important to emphasize that, although the purpose of a ModulatorySignal is to modify the functioning of a
`Mechanism` or a `MappingProjection`, it does this indirectly by modifying a State that determines the input or
output of a Mechanism, or the parameters of a Mechanism or Projection's `function`, rather than directly modifying
the function of the Mechanism or Projection itself.  This is shown in the following figure, and described in greater
detail under `ModulatorySignal_Implementation`.

.. _ModulatorySignal_Anatomy_Figure:

**Anatomy of Modulation**

.. figure:: _static/Modulation_Anatomy_fig.svg
   :alt: Modulation
   :scale: 150 %

   **Three types of Modulatory Components and the States they modulate**. The default `type of modulation
   <ModulatorySignal_Types>` for each type of ModulatorySignal, and the default Function and modulated parameter of
   its recipient State are listed in the `table <ModulatorySignal_Table>` above. Note that the `ControlMechanism`
   and `ControlSignal <ControlSignal>` are shown in the figure modulating the `ParameterState` of a Mechanism;
   however, like Gating components, they can also be used to modulate `InputStates <InputState>` and `OutputStates
   <OutputState>`. The `figure <ModulatorySignal_Detail_Figure>` below shows a detailed view of how ModulatorySignals
   modulate the parameters of a State's `function <State_Base.function>`.


.. _ModulatorySignal_Types:

Types of Modulation
^^^^^^^^^^^^^^^^^^^

The `modulation <ModulatorySignal.modulation>` attribute of a ModulatorySignal determines the way in which it
modulates the `value <State_Base.value>` of a `State`, by specifying which paramter of the State's `function
<State_Base.function>`that it modifies (see `figure <ModulatorySignal_Detail_Figure>` below).  This is specified
in a tuple containing the State and the name of the parameter to be modified (see `example <EXAMPLE??> below).
Alternatively, there are four keywords that can be used in place of the parameter's name, that specify the two most
commonly used types of modulation, and allow two other types:

  * *MULTPLICATIVE_PARAM* - assign the `value <ModulatorySignal.value>` of the ModulatorySignal to the parameter of
    the State's `function <State_Base.function>` specified as its `multiplicative_param <Function_Modulatory_Params>`.
    For example, if the State's `function <State_Base.function>` is `Linear` (the default for most States), then
    the ModulatorySignal's `value <ModulatorySignal.value>` is assigned to the function's `slope <Linear.slope>`
    parameter (it's multiplicative_param), thus multiplying the State's `variable <State_Base.variable>` by that
    amount each time the State is executed, and assigning the result as the State's `value <State_Base.value>`.

  * *ADDITIVE_PARAM* - assign the `value <ModulatorySignal.value>` of the ModulatorySignal to the parameter of the
    State's `function <State_Base.function>` specified as its `additive_param <Function_Modulatory_Params>`. For
    example, if the State's `function <State_Base.function>` is `Linear` (the default for most States), then the
    ModulatorySignal's `value <ModulatorySignal.value>` is assigned to the function's `intercept <Linear.intercept>`
    parameter (it's additive_param), thus adding that value to the State's `variable <State_Base.variable>` each
    time the State is executed, and assigning the result as the State's `value <State_Base.value>`.

  * *OVERRIDE* - assign the `value <ModulatorySignal.value>` of the ModulatorySignal directly to the State's
    `value <State_Base.value>`; in effect, this bypasses the State's `function <State_Base.function>`. Note that
    this can be specified for **only one** `ModulatorySignal` that modulates a given State (see `below
    <ModulatorySignal_Multiple>` for additional details).

  * *DISABLE* - suppresses the modulatory effect of the ModulatorySignal;  the State's `function <State_Base.function>`
    will operate as if it did not receive a `ModulatoryProjection <ModulatoryProjection>` from that ModulatorySignal.

   .. note:

      the *MULTPLICATIVE_PARAM* and *ADDITIVE_PARAM* keywords can be used only with `Functions <Function>` that
      specify a `multiplicative_param and/or additive_param <Function_Modulatory_Params>`, respectively.

COMMENT:
FOR DEVELOPERS:  the MULTPLICATIVE_PARAM and ADDITIVE_PARAM options above are keywords for aliases to the relevant
parameters of a given Function, declared in its Parameters subclass declaration of the Function's declaration.
COMMENT

The default type of modulation for `ControlSignals <ControlSignal>` and `GatingSignals <GatingSignal>` is
*MULTIPLICATIVE*.  The default for `LearningSignals <LearningSignal>` is *ADDITIVE* (which additively modifies the
`value <LearningSignal.value>` of the LearningSignal (i.e., the weight changes computed by the `LearningMechanism`)
to the State's `variable <State_Base.variable>` (i.e., the current weight `matrix <MappingProjection.matrix>` for
the `MappingProjection` being learned).


.. _ModulatorySignal_Implementation:

Implementation of Modulation
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Although the `purpose of a ModulatorySignal <ModulatorySignal_Uses>` is to modify the operation of
a `Mechanism` or `MappingProjection`, it does not do this *directly*; rather, it does it by way of a `State` that
modulates the input, function parameter, or output of the Mechanism or MappingProjeciton to be regulated. More
specifically, a ModulatorySignal modulates the `function <State_Base.function>` of a *State* responsible for
generating those values, by modifying a parameter of that State's `function <State_Base.function>`.  This is shown
in the figure below:

.. _ModulatorySignal_Detail_Figure:

**Detailed View of Modulation**

.. figure:: _static/Modulation_Detail_fig.svg
   :alt: Modulation_Detail
   :scale: 150 %

   A ModulatorySignal modulates the `value <State_Base.value>` of a State either by modifying a parameter of the
   State's `function <State_Base.function>`, or assigining the `value <State_Base.value>` of the State directly.
   This is determined by the ModulatorySignal's `modulation <ModulatorySignal.modulation>` attribute.  That can be
   assigned either the name of a parameter of the State's `function <State_Base.function>`, or a keyword that
   specifies a standard form of modulation.  The keywords *MULTIPLICATIVE* and *ADDITIVE* specify that the `value
   <ModulatorySignal.value>` of the ModulatorySignal be assigned to the `multiplicative_param or `additive_param
   <Function_Modulatory_Params>` of the State's function, respectively;  *OVERRIDE* specifies that the
   ModulatorySignal's `value <ModulatorySignal.value>` be assigned directly as the State's `value
   <State_Base.value>`, in effect bypassing the State's `function <State_Base.function>` (see
   `ModulatorySignal_Types` for additional details).

Though this implementaton of modulation is indirect, it provides a standard for all forms of modulation, as well as
considerable flexibility in the adaptive regulation of Components within a `Composition` (see
`ModulatorySignal_Types` below).

The types of States modulated by each type of ModulatorySignal are summarized in `ModulatorySignal_Uses`,
and the accompanying `table <ModulatorySignal_Table>` and `figure <ModulatorySignal_Anatomy_Figure>`.

.. _ModulatorySignal_Multiple:

Any `modulable <Parameter.modulable>` parameter of a State's `function <State_Base.function>` can modulated,
and different parameters of the same `function <State_Base.function>` of a State can be modulated by different
ModulatorySignals. The same parameter can also be modulated by more than on ModulatorySignal. If more than one
ModulatorySignal modulates the same parameter of a State's `function <State_Base.function>`, then that parameter's
`modulation_combine_function <Parameter.modulation_combine_function>` attribute determines how the `value
<ModulatorySignal.value>`\\s of the different ModulatorySignals are combined.  By default, the product of their
vaues is used.  However, if *OVERRIDE* is specfied as the type of `modulation <ModulatorySignal.modulation>` for one
of them, then that ModulatorySignal's  `value <ModulatorySignal.value>` is assigned directly as the State's `value
<State_Base.value>`, and the others are all ignored.  Only one ModulatorySignal specified as *OVERRIDE* can modulate
a given parameter;  if there is more than, then an error is generated.

.. _ModulatorySignal_Execution:

Execution
---------

ModulatorySignals cannot be executed directly.  This done when the `AdaptiveMechanism <AdaptiveMechanism>` to
which they belong is executed. When a ModulatorySignal is executed, it calculates its `value <ModulatorySignal.value>`,
which is then assigned as the `variable <ModulatoryProjection_Base.variable>` of the `ModulatoryProjections
<ModulatoryProjection>` listed in its `efferents <ModulatorySignal.efferents>` attribute.
When those Projections execute, they convey the ModulatorySignal's `value <ModulatorySignal.value>` to the `function
<State_Base.function>` of the `State <State>` to which they project.  The State's `function <State_Base.function>`
then uses that value in determining value of the parameter designated by the `modulation <ModulatorySignal.modulation>`
attribute of the ModulatorySignal when the State's `value <State_Baselvalue>` is updated.

COMMENT:

# FIX: 9/3/19 -- REWORK AND ADD EXAMPLE HERE

For example, consider a `ControlSignal` that modulates the `bias <Logistic.bias>` parameter of a `Logistic` Function
used by a `TransferMechanism`, and assume that the `ParameterState` for the bias parameter (to which the ControlSignal
projects) uses a `Linear` function (the default for a ParameterState) to set the `value <ParameterState.value>` of
that parameter. If the `modulation  <ModulatorySignal.modulation>` attribute of the `ControlSignal` is *MULTIPLICATIVE*
then, when the TransferMechanism's `Logistic` `function <TransferMechanism.function>` is executed, the `function
<ParameterState.function>` of the ParameterState that sets the value of the `Logistic` Function's `bias <Logistic.bias>`
parameter is executed;  that is a `Linear` Function, that uses the ControlSignal's `value <ControlSignal.value>` as
its `slope <Linear.slope>` parameter.  Thus, the effect is that the ControlSignal's `value <ControlSignal.value>` is
multiplied by the base value of the `bias <Logistic.bias>` parameter, before that is used by the TransferMechanism's
`Logistic` Function.  Thus, the `value <ControlSignal.value>` of the ControlSignal modulates the `bias
<Logistic.bias>` parameter of the `Logistic` Function when the TransferMechanism's `function
<TransferMechanism.function>` is executed (see `State Execution <State_Execution>` for additional details).

COMMENT

.. note::
   The change in the value of a `State <State>` in response to a ModulatorySignal does not occur until the Mechanism to
   which the state belongs is next executed; see :ref:`Lazy Evaluation <LINK>` for an explanation of "lazy" updating).

.. _ModulatorySignal_Class_Reference:

COMMENT:

Examples
--------

MOVE THESE TO SPECIFIC SUBCLASSES? AND REFERENCE THOSE HERE
FIX: EXAMPLE OF FULL SPECIFIATION (BY STATE AND STATE'S FUCNTION'S PARAMETER NAME)
The following example uses a parameter's name to specify
    >>> my_mech = ProcessingMechanism(function=Logistic)
    >>> ctl_mech = ControlMechanism(monitor_for_control=my_mech,
    ...                             control_signals=ControlSignal(modulates=my_mech.parameter_states[GAIN],
    ...                                                           modulation=SLOPE))

FIX: EXAMPLE OF SPECIFIATION OF CONTROLSIGNAL WITH MECHANISM AND STATE'S PARAMETER NAME

FIX: EXAMPLE OF SPECIFIATION BY CONTROLSIGNAL WITH MECHANISM AND MECHANISM'S PARAMETER NAME

MENTION STATE-SPECIFIC CONVENIENCE METHODS

FIX: EXAMPLE OF CONTROL SIGNAL MODULATION OF INPUT STATE


.  For
example, the `TransferWithCosts` `Function` defines keywords for `modulating the parameters of its cost functions
<TransferWithCosts_Modulation_of_Cost_Params>`.
A ControlMechanism can even modulate the parameters of another
ControlMechanism, or its ControlSignals.  For example, in the following, ``ctl_mech_A`` modulates the `intensity_cost
<ControlSignal.intensity_cost>` parameter of ``ctl_mech``\\'s ControlSignal::

    >>> my_mech = ProcessingMechanism()
    >>> ctl_mech_A = ControlMechanism(monitor_for_control=my_mech,
    ...                               control_signals=ControlSignal(modulates=(SLOPE,my_mech),
    >>>                                                              cost_options = CostFunctions.INTENSITY))
    >>> ctl_mech_B = ControlMechanism(monitor_for_control=my_mech,
    ...                               control_signals=ControlSignal(modulates=ctl_mech_A.control_signals[0],
    ...                                                             modulation=INTENSITY_COST_FCT_MULTIPLICATIVE_PARAM))

FIX: EXAMPLE HERE
     THEN, AFTER EXAMPLE: MODULATORYSIGNAL'S VALUE IS ASSIGNED TO THE SPECIFIED PARAMETER.

FIX: PUT THIS IN ModulatoryMechanism AND CROSS REFERENCE HERE? OR VICE VERSA?
EXAMPLE OF MIXED MODULATORY SIGNALS IN MODULATORY MECHANISM

        m = ProcessingMechanism(function=Logistic)
        c = ModulatoryMechanism(
                modulatory_signals=[
                    ControlSignal(name="CS1", modulates=(GAIN, m)),
                    GatingSignal(name="GS", modulates=m),
                    ControlSignal(name="CS2", modulates=(BIAS, m)),
                ]
        )


COMMENT

Class Reference
---------------
"""

from psyneulink.core.components.component import component_keywords
from psyneulink.core.components.states.outputstate import OutputState
from psyneulink.core.components.states.state import State_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import MAYBE, MECHANISM, MODULATION, MODULATORY_SIGNAL, VARIABLE, PROJECTIONS
from psyneulink.core.globals.defaults import defaultModulatoryAllocation
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel

__all__ = [
    'modulatory_signal_keywords', 'ModulatorySignal', 'ModulatorySignalError',
]


def _is_modulatory_spec(spec, include_matrix_spec=True):
    from psyneulink.core.components.mechanisms.adaptive.learning.learningmechanism import _is_learning_spec
    from psyneulink.core.components.mechanisms.adaptive.control.controlmechanism import _is_control_spec
    from psyneulink.core.components.mechanisms.adaptive.gating.gatingmechanism import _is_gating_spec

    if (_is_learning_spec(spec, include_matrix_spec=include_matrix_spec)
        or _is_control_spec(spec)
        or _is_gating_spec(spec)
        ):
        return True
    else:
        return False


class ModulatorySignalError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)

modulatory_signal_keywords = {MECHANISM, MODULATION}
modulatory_signal_keywords.update(component_keywords)


class ModulatorySignal(OutputState):
    """
    ModulatorySignal(                                  \
        owner,                                         \
        default_allocation=defaultModulatoryAllocation \
        function=LinearCombination(operation=SUM),     \
        modulation=MULTIPLICATIVE                      \
        projections=None,                              \
        params=None,                                   \
        name=None,                                     \
        prefs=None)

    Subclass of `OutputState` used by an `AdaptiveMechanism <AdaptiveMechanism>` to modulate the value
    of one more `States <State>`.

    .. note::
       ModulatorySignal is an abstract class and should NEVER be instantiated by a call to its constructor.
       It should be instantiated using the constructor for a `subclass <ModulatorySignal_Subtypes>`.

    COMMENT:

        Description
        -----------
            The ModulatorySignal class is a subtype of the OutputState class in the State category of Component,
            It is used primarily as the sender for GatingProjections
            Its FUNCTION updates its value:
                note:  currently, this is the identity function, that simply maps variable to self.value

        Class attributes:
            + componentType (str) = GATING_SIGNAL
            + paramClassDefaults (dict)
                + FUNCTION (LinearCombination)
                + FUNCTION_PARAMS (Modulation.MULTIPLY)

        Class methods:
            function (executes function specified in params[FUNCTION];  default: Linear

        StateRegistry
        -------------
            All OutputStates are registered in StateRegistry, which maintains an entry for the subclass,
              a count for all instances of it, and a dictionary of those instances
    COMMENT


    Arguments
    ---------

    owner : ModulatoryMechanism
        specifies the `GatingMechanism` to which to assign the ModulatorySignal.

    default_allocation : scalar : defaultModulatoryAllocation
        specifies the default template and value used for `variable <ModulatorySignal.variable>`.

    function : Function or method : default Linear
        specifies the function used to determine the value of the ModulatorySignal from the value of its
        `owner <GatingMechanism.owner>`.

    modulation : ModulationParam : default MULTIPLICATIVE
        specifies the type of modulation the ModulatorySignal uses to determine the value of the State(s) it modulates.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterState_Specification>` that can be used to specify the parameters for
        the ControlSignal and/or a custom function and its parameters. Values specified for parameters in the dictionary
        override any assigned to those parameters in arguments of the constructor.

    name : str : default see `name <ModulatorySignal.name>`
        specifies the name of the ModulatorySignal.

    prefs : PreferenceSet or specification dict : default State.classPreferences
        specifies the `PreferenceSet` for the LearningSignal; see `prefs <ControlSignal.prefs>` for details.


    Attributes
    ----------

    owner : AdaptiveMechanism
        the `AdaptiveMechanism <AdaptiveMechanism>` to which the ModulatorySignal belongs.

    variable : scalar, list or np.ndarray
        same as `allocation <ModulatorySignal.allocation>`.

    allocation : float
        value assigned by the ModulatorySignal's `owner <ControlSignal.owner>`, and used as the `variable
        <ModulatorySignal.variable>` of its `function <ModulatorySignal.function>` to determine the ModulatorySignal's
        `ModulatorySignal.intensity`.
    COMMENT:
    FOR DEVELOPERS:  Implemented as an alias of the ModulatorySignal's variable Parameter
    COMMENT

    function : TransferFunction
        provides the ModulatorySignal's `value <ModulatorySignal.value>`; the default is an identity function that
        assigns `variable <ModulatorySignal.variable>` as ModulatorySignal's `value <ModulatorySignal.value>`.

    value : number, list or np.ndarray
        result of `function <ModulatorySignal.function>`, used to determine the `value <State_Base.value>` of the
        State(s) being modulated.

    modulation : ModulationParam
        determines how the output of the ModulatorySignal is used to modulate the value of the state(s) being modulated.

    efferents : [List[GatingProjection]]
        a list of the `ModulatoryProjections <ModulatoryProjection>` assigned to the ModulatorySignal.

    name : str
        the name of the ModulatorySignal. If the ModulatorySignal's `initialization has been deferred
        <State_Deferred_Initialization>`, it is assigned a temporary name (indicating its deferred initialization
        status) until initialization is completed, at which time it is assigned its designated name.  If that is the
        name of an existing ModulatorySignal, it is appended with an indexed suffix, incremented for each State with
        the same base name (see `Naming`). If the name is not  specified in the **name** argument of its constructor,
        a default name is assigned as follows; if the ModulatorySignal has:

        * no projections (which are used to name it) -- the name of its class is used, with an index that is
          incremented for each ModulatorySignal with a default named assigned to its `owner <ModulatorySignal.owner>`;

        * one `ModulatoryProjection <ModulatoryProjction>` -- the following template is used:
          "<target Mechanism name> <target State name> <ModulatorySignal type name>"
          (for example, ``'Decision[drift_rate] ControlSignal'``, or ``'Input Layer[InputState-0] GatingSignal'``);

        * multiple ModulatoryProjections, all to States of the same Mechanism -- the following template is used:
          "<target Mechanism name> (<target State name>,...) <ModulatorySignal type name>"
          (for example, ``Decision (drift_rate, threshold) ControlSignal``, or
          ``'Input Layer[InputState-0, InputState-1] GatingSignal'``);

        * multiple ModulatoryProjections to States of different Mechanisms -- the following template is used:
          "<owner Mechanism's name> divergent <ModulatorySignal type name>"
          (for example, ``'ControlMechanism divergent ControlSignal'`` or ``'GatingMechanism divergent GatingSignal'``).

        .. note::
            Unlike other PsyNeuLink components, State names are "scoped" within a Mechanism, meaning that States with
            the same name are permitted in different Mechanisms.  However, they are *not* permitted in the same
            Mechanism: States within a Mechanism with the same base name are appended an index in the order of their
            creation.

    prefs : PreferenceSet or specification dict
        the `PreferenceSet` for the ModulatorySignal; if it is not specified in the **prefs** argument of the
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see :doc:`PreferenceSet
        <LINK>` for details).

    """

    componentType = MODULATORY_SIGNAL
    # paramsType = OUTPUT_STATE_PARAMS

    class Parameters(OutputState.Parameters):
        """
            Attributes
            ----------

                modulation
                    see `modulation <ModulatorySignal.modulation>`

                    :default value: None
                    :type:

        """
        modulation = None

    stateAttributes =  OutputState.stateAttributes | {MODULATION}

    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TypeDefaultPreferences
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     kwPreferenceSetName: 'OutputStateCustomClassPreferences',
    #     kp<pref>: <setting>...}

    paramClassDefaults = State_Base.paramClassDefaults.copy()

    def __init__(self,
                 owner=None,
                 size=None,
                 reference_value=None,
                 default_allocation=defaultModulatoryAllocation,
                 function=None,
                 modulates=None,
                 modulation=None,
                 index=None,
                 assign=None,
                 params=None,
                 name=None,
                 prefs=None,
                 **kwargs):

        if kwargs:
            if VARIABLE in kwargs:
                default_allocation = kwargs.pop(VARIABLE, default_allocation)
            if PROJECTIONS in kwargs:
                modulates = kwargs.pop(PROJECTIONS, modulates)

        # Deferred initialization
        # if self.initialization_status & (ContextFlags.DEFERRED_INIT | ContextFlags.INITIALIZING):
        if self.initialization_status & ContextFlags.DEFERRED_INIT:
            # If init was deferred, it may have been because owner was not yet known (see OutputState.__init__),
            #   and so modulation hasn't had a chance to be assigned to the owner's value
            #   (i.e., if it was not specified in the constructor), so do it now;
            #   however modulation has already been assigned to params, so need to assign it there
            params[MODULATION] = self.modulation or owner.modulation

        # Standard initialization
        else:
            # Assign args to params and functionParams dicts
            params = self._assign_args_to_param_dicts(params=params,
                                                      modulation=modulation)

        super().__init__(owner=owner,
                         reference_value=reference_value,
                         variable=default_allocation,
                         size=size,
                         projections=modulates,
                         index=index,
                         assign=assign,
                         function=function,
                         params=params,
                         name=name,
                         prefs=prefs,
                         **kwargs)

        if self.initialization_status == ContextFlags.INITIALIZED:
            self._assign_default_state_name()

    def _instantiate_attributes_after_function(self, context=None):
        # If owner is specified but modulation has not been specified, assign to owner's value

        super()._instantiate_attributes_after_function(context=context)
        if self.owner and self.modulation is None:
            self.modulation = self.owner.modulation


    def _instantiate_projections(self, projections, context=None):
        """Instantiate Projections specified in PROJECTIONS entry of params arg of State's constructor

        Specification should be an existing ModulatoryProjection, or a receiver Mechanism or State
        Disallow any other specifications (including PathwayProjections)
        Call _instantiate_projection_from_state to assign ModulatoryProjections to .efferents

        """
       # IMPLEMENTATION NOTE: THIS SHOULD BE MOVED TO COMPOSITION ONCE THAT IS IMPLEMENTED
        for receiver_spec in projections:
            projection = self._instantiate_projection_from_state(projection_spec=type(self),
                                                                 receiver=receiver_spec,
                                                                 # MODIFIED 8/12/19 NEW: [JDC] - MODIFIED FEEDBACK
                                                                 # feedback=True,
                                                                 feedback=MAYBE,
                                                                 # MODIFIED 8/12/19 END
                                                                 context=context)
            # Projection might be None if it was duplicate
            if projection:
                projection._assign_default_projection_name(state=self)

    def _assign_default_state_name(self, context=None):

        # If the name is not a default name for the class,
        #    or the ModulatorySignal has no projections (which are used to name it)
        #    then return
        if (not (self.name is self.__class__.__name__
                 or self.__class__.__name__ + '-' in self.name) or
                    len(self.efferents)==0):
            return self.name

        # Construct default name
        receiver_names = []
        receiver_owner_names = []
        receiver_owner_receiver_names = []
        class_name = self.__class__.__name__

        for projection in self.efferents:
            receiver = projection.receiver
            receiver_name = receiver.name
            receiver_owner_name = receiver.owner.name
            receiver_names.append(receiver_name)
            receiver_owner_names.append(receiver_owner_name)
            receiver_owner_receiver_names.append("{}[{}]".format(receiver_owner_name, receiver_name))

        # Only one ModulatoryProjection: "<target mech> <State.name> <ModulatorySignal>"
        # (e.g., "Decision drift_rate ControlSignal", or "Input Layer InputState-0 GatingSignal")
        if len(receiver_owner_receiver_names) == 1:
            default_name = receiver_owner_receiver_names[0] + " " + class_name

        # Multiple ModulatoryProjections all for same mech: "<target mech> (<State.name>,...) <ModulatorySignal>"
        # (e.g., "Decision (drift_rate, threshold) ControlSignal" or
        #        "InputLayer (InputState-0, InputState-0) ControlSignal")
        elif all(name is receiver_owner_names[0] for name in receiver_owner_names):
            default_name = "{}[{}] {}".format(receiver_owner_names[0], ", ".join(receiver_names), class_name)

        # Mult ModulatoryProjections for diff mechs: "<owner mech> divergent <ModulatorySignal>"
        # (e.g., "EVC divergent ControlSignal", or "GatingMechanism divergent GatingSignal")
        else:
            default_name = self.name + " divergent " + class_name

        self.name = default_name

        return self.name



import numpy as np
import psyneulink as pnl
import pytest

from collections import OrderedDict

import psyneulink.core.components.functions.transferfunctions
from psyneulink.core.globals.keywords import ALLOCATION_SAMPLES, PROJECTIONS

class TestLog:

    def test_log(self):

        T_1 = pnl.TransferMechanism(name='log_test_T_1', size=2)
        T_2 = pnl.TransferMechanism(name='log_test_T_2', size=2)
        PS = pnl.Process(name='log_test_PS', pathway=[T_1, T_2])
        PJ = T_2.path_afferents[0]

        assert T_1.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'OFF',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T_2.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'OFF',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_value': 'OFF',
            'has_initializers': 'OFF',
            'func_variable': 'OFF',
            'matrix': 'OFF',
            'mod_matrix': 'OFF',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        T_1.set_log_conditions('mod_noise')
        T_1.set_log_conditions(pnl.RESULTS)
        PJ.set_log_conditions('mod_matrix')

        assert T_1.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'EXECUTION',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'EXECUTION',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T_2.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'OFF',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'matrix': 'OFF',
            'mod_matrix': 'EXECUTION',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        PS.execute()
        PS.execute()
        PS.execute()

        assert T_1.logged_items == {'RESULTS': 'EXECUTION', 'mod_noise': 'EXECUTION'}
        assert PJ.logged_items == {'mod_matrix': 'EXECUTION'}

        T_1.log.print_entries(contexts=PS)

        # assert T_1.log.print_entries() ==
        # # Log for mech_A:
        # #
        # # Index     Variable:                                          Context                                                                  Value
        # # 0         'RESULTS'.........................................' EXECUTING  PROCESS Process-0'.......................................    0.0
        # # 1         'RESULTS'.........................................' EXECUTING  PROCESS Process-0'.......................................    0.0
        # #
        # #
        # # 0         'noise'...........................................' EXECUTING  PROCESS Process-0'.......................................    0.0
        # # 1         'noise'...........................................' EXECUTING  PROCESS Process-0'.......................................    0.0
        #
        # assert T_2.log.print_entries() ==
        # # Log for mech_A:
        # #
        # # Index     Variable:                                          Context                                                                  Value
        # # 0         'RESULTS'.........................................' EXECUTING  PROCESS Process-0'.......................................    0.0
        # # 1         'RESULTS'.........................................' EXECUTING  PROCESS Process-0'.......................................    0.0
        # #
        # #
        # # 0         'noise'...........................................' EXECUTING  PROCESS Process-0'.......................................    0.0
        # # 1         'noise'...........................................' EXECUTING  PROCESS Process-0'.......................................    0.0

        T_1_csv = T_1.log.csv(entries=['mod_noise', 'RESULTS'], owner_name=False, quotes=None)
        print(T_1_csv)
        assert T_1_csv == \
            "'Execution Context', 'Data'\n" \
            + "'{0}', \'Index\', \'mod_noise\', \'RESULTS\'\n".format(PS.default_execution_id) \
            + ", 0, 0.0, 0.0 0.0\n" \
            + ", 1, 0.0, 0.0 0.0\n" \
            + ", 2, 0.0, 0.0 0.0\n"

        assert PJ.log.csv(entries='mod_matrix', owner_name=True, quotes=True) == \
            "'Execution Context', 'Data'\n" \
            + "'{0}', \'Index\', \'MappingProjection from log_test_T_1 to log_test_T_2[mod_matrix]\'\n".format(PS.default_execution_id) \
            + ", \'0\', \'1.0 0.0\' \'0.0 1.0\'\n" \
            + ", \'1\', \'1.0 0.0\' \'0.0 1.0\'\n" \
            + ", \'2\', \'1.0 0.0\' \'0.0 1.0\'\n"

        result = T_1.log.nparray(entries=['mod_noise', 'RESULTS'], header=False, owner_name=True)
        assert result[0] == PS.default_execution_id
        np.testing.assert_array_equal(result[1][0],
                                      np.array([[[0], [1], [2]],
                                                [[ 0.], [ 0.], [ 0.]],
                                                [[ 0.,  0.], [ 0.,  0.],[ 0., 0.]]]))

    def test_log_initialization(self):
        T = pnl.TransferMechanism(
                prefs={pnl.LOG_PREF: pnl.PreferenceEntry(pnl.LogCondition.INITIALIZATION, pnl.PreferenceLevel.INSTANCE)}
        )
        assert T.logged_items == {'value': 'INITIALIZATION'}

    def test_log_dictionary_without_time(self):

        T1 = pnl.TransferMechanism(name='log_test_T1',
                                    size=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                    size=2)
        PS = pnl.Process(name='log_test_PS', pathway=[T1, T2])
        PJ = T2.path_afferents[0]

        assert T1.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'OFF',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T2.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'OFF',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'matrix': 'OFF',
            'mod_matrix': 'OFF',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULTS)
        T1.set_log_conditions(pnl.VALUE)
        PJ.set_log_conditions('mod_matrix')
        T2.set_log_conditions('mod_slope')
        T2.set_log_conditions(pnl.RESULTS)
        T2.set_log_conditions(pnl.VALUE)

        assert T1.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'EXECUTION',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }
        assert T2.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'EXECUTION',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }
        assert PJ.loggable_items == {
            'exponent': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_matrix': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'has_initializers': 'OFF',
            'matrix': 'OFF',
            'mod_matrix': 'EXECUTION',
            'value': 'OFF',
            'variable': 'OFF',
            'weight': 'OFF'
        }

        PS.execute([1.0, 2.0])
        PS.execute([3.0, 4.0])
        PS.execute([5.0, 6.0])

        assert T1.logged_items == {'RESULTS': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}
        assert T2.logged_items == {'RESULTS': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}
        assert PJ.logged_items == {'mod_matrix': 'EXECUTION'}

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULTS'])

        expected_values_T1 = [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]
        expected_slopes_T1 = [[1.0], [1.0], [1.0]]
        expected_results_T1 = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]

        assert list(log_dict_T1.keys()) == [PS.default_execution_id]

        assert np.allclose(expected_values_T1, log_dict_T1[PS.default_execution_id]['value'])
        assert np.allclose(expected_slopes_T1, log_dict_T1[PS.default_execution_id]['mod_slope'])
        assert np.allclose(expected_results_T1, log_dict_T1[PS.default_execution_id]['RESULTS'])

        assert list(log_dict_T1[PS.default_execution_id].keys()) == ['Index', 'value', 'mod_slope', 'RESULTS']

        log_dict_T1_reorder = T1.log.nparray_dictionary(entries=['mod_slope', 'value', 'RESULTS'])

        assert list(log_dict_T1_reorder[PS.default_execution_id].keys()) == ['Index', 'mod_slope', 'value', 'RESULTS']

    def test_run_resets(self):
        import psyneulink as pnl
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   size=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   size=2)
        PS = pnl.Process(name='log_test_PS', pathway=[T1, T2])
        SYS = pnl.System(name='log_test_SYS', processes=[PS])
        T1.set_log_conditions('mod_slope')
        T2.set_log_conditions('mod_slope')
        SYS.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        log_array_T1 = T1.log.nparray()
        log_array_T2 = T2.log.nparray()

        SYS.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        log_array_T1_second_run = T1.log.nparray()
        log_array_T2_second_run = T2.log.nparray()

    def test_log_dictionary_with_time(self):

        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   size=2)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=psyneulink.core.components.functions.transferfunctions.Linear(slope=2.0),
                                   size=2)
        PS = pnl.Process(name='log_test_PS', pathway=[T1, T2])
        SYS = pnl.System(name='log_test_SYS', processes=[PS])

        assert T1.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'OFF',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }
        assert T2.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'OFF',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'OFF',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'OFF',
            'variable': 'OFF'
        }

        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULTS)
        T1.set_log_conditions(pnl.VALUE)

        assert T1.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'EXECUTION',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }

        T2.set_log_conditions('mod_slope')
        T2.set_log_conditions(pnl.RESULTS)
        T2.set_log_conditions(pnl.VALUE)

        assert T2.loggable_items == {
            'InputState-0': 'OFF',
            'RESULTS': 'EXECUTION',
            'clip': 'OFF',
            'convergence_criterion': 'OFF',
            'func_additive_param': 'OFF',
            'func_bounds': 'OFF',
            'func_has_initializers': 'OFF',
            'func_intercept': 'OFF',
            'func_multiplicative_param': 'OFF',
            'func_slope': 'OFF',
            'func_value': 'OFF',
            'func_variable': 'OFF',
            'function': 'OFF',
            'has_initializers': 'OFF',
            'initial_value': 'OFF',
            'integration_rate': 'OFF',
            'integrator_function_value': 'OFF',
            'integrator_mode': 'OFF',
            'max_passes': 'OFF',
            'mod_convergence_criterion': 'OFF',
            'mod_integration_rate': 'OFF',
            'mod_intercept': 'OFF',
            'mod_noise': 'OFF',
            'mod_slope': 'EXECUTION',
            'noise': 'OFF',
            'previous_value': 'OFF',
            'value': 'EXECUTION',
            'variable': 'OFF'
        }

        # RUN ZERO  |  TRIALS ZERO, ONE, TWO ----------------------------------

        SYS.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        assert T1.logged_items == {'RESULTS': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}
        assert T2.logged_items == {'RESULTS': 'EXECUTION',
                                   'mod_slope': 'EXECUTION',
                                   'value': 'EXECUTION'}

        # T1 log after zero-th run -------------------------------------------

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULTS'])

        expected_run_T1 = [[0], [0], [0]]
        expected_trial_T1 = [[0], [1], [2]]
        expected_time_step_T1 = [[0], [0], [0]]
        expected_values_T1 = [[[1.0, 2.0]], [[3.0, 4.0]], [[5.0, 6.0]]]
        expected_slopes_T1 = [[1.0], [1.0], [1.0]]
        expected_results_T1 = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]

        assert list(log_dict_T1.keys()) == [SYS.default_execution_id]
        sys_log_dict = log_dict_T1[SYS.default_execution_id]

        assert np.allclose(expected_run_T1, sys_log_dict['Run'])
        assert np.allclose(expected_trial_T1, sys_log_dict['Trial'])
        assert np.allclose(expected_time_step_T1, sys_log_dict['Time_step'])
        assert np.allclose(expected_values_T1, sys_log_dict['value'])
        assert np.allclose(expected_slopes_T1, sys_log_dict['mod_slope'])
        assert np.allclose(expected_results_T1, sys_log_dict['RESULTS'])

        # T2 log after zero-th run --------------------------------------------

        log_dict_T2 = T2.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULTS'])

        expected_run_T2 = [[0], [0], [0]]
        expected_trial_T2 = [[0], [1], [2]]
        expected_time_step_T2 = [[1], [1], [1]]
        expected_values_T2 = [[[2.0, 4.0]], [[6.0, 8.0]], [[10.0, 12.0]]]
        expected_slopes_T2 = [[2.0], [2.0], [2.0]]
        expected_results_T2 = [[2.0, 4.0], [6.0, 8.0], [10.0, 12.0]]

        assert list(log_dict_T2.keys()) == [SYS.default_execution_id]
        sys_log_dict = log_dict_T2[SYS.default_execution_id]

        assert np.allclose(expected_run_T2, sys_log_dict['Run'])
        assert np.allclose(expected_trial_T2, sys_log_dict['Trial'])
        assert np.allclose(expected_time_step_T2, sys_log_dict['Time_step'])
        assert np.allclose(expected_values_T2, sys_log_dict['value'])
        assert np.allclose(expected_slopes_T2, sys_log_dict['mod_slope'])
        assert np.allclose(expected_results_T2, sys_log_dict['RESULTS'])

        # RUN ONE  |  TRIALS ZERO, ONE, TWO -------------------------------------

        SYS.run(inputs={T1: [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]})

        # T1 log after first run -------------------------------------------

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULTS'])

        assert list(log_dict_T1.keys()) == [SYS.default_execution_id]
        sys_log_dict = log_dict_T1[SYS.default_execution_id]

        # expected_run_T1_2 = [[1], [1], [1]]
        expected_run_T1_2 = [[0], [0], [0]] + expected_run_T1
        expected_trial_T1_2 = [[0], [1], [2]] + expected_trial_T1
        expected_time_step_T1_2 = [[0], [0], [0]] + expected_time_step_T1
        expected_values_T1_2 = expected_values_T1 + expected_values_T1
        expected_slopes_T1_2 = expected_slopes_T1 + expected_slopes_T1
        expected_results_T1_2 = expected_results_T1 + expected_results_T1

        # assert np.allclose(expected_run_T1_2, sys_log_dict['Run'])
        # assert np.allclose(expected_trial_T1_2, sys_log_dict['Trial'])
        # assert np.allclose(expected_time_step_T1_2, sys_log_dict['Time_step'])
        assert np.allclose(expected_values_T1_2, sys_log_dict['value'])
        assert np.allclose(expected_slopes_T1_2, sys_log_dict['mod_slope'])
        assert np.allclose(expected_results_T1_2, sys_log_dict['RESULTS'])

        # T2 log after first run -------------------------------------------

        log_dict_T2_2 = T2.log.nparray_dictionary(entries=['value', 'mod_slope', 'RESULTS'])

        assert list(log_dict_T2_2.keys()) == [SYS.default_execution_id]
        sys_log_dict = log_dict_T2_2[SYS.default_execution_id]

        expected_run_T2_2 = [[0], [0], [0]] + expected_run_T2
        expected_trial_T2_2 = [[0], [1], [2]] + expected_trial_T2
        expected_time_step_T2_2 = [[1], [1], [1]] + expected_time_step_T2
        expected_values_T2_2 = [[[2.0, 4.0]], [[6.0, 8.0]], [[10.0, 12.0]]] + expected_values_T2
        expected_slopes_T2_2 = [[2.0], [2.0], [2.0]] + expected_slopes_T2
        expected_results_T2_2 = [[2.0, 4.0], [6.0, 8.0], [10.0, 12.0]] + expected_results_T2

        # assert np.allclose(expected_run_T2_2, sys_log_dict['Run'])
        # assert np.allclose(expected_trial_T2_2, sys_log_dict['Trial'])
        # assert np.allclose(expected_time_step_T2_2, sys_log_dict['Time_step'])
        assert np.allclose(expected_values_T2_2, sys_log_dict['value'])
        assert np.allclose(expected_slopes_T2_2, sys_log_dict['mod_slope'])
        assert np.allclose(expected_results_T2_2, sys_log_dict['RESULTS'])

    def test_log_dictionary_with_scheduler(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.5)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=psyneulink.core.components.functions.transferfunctions.Linear(slope=6.0))
        PS = pnl.Process(name='log_test_PS', pathway=[T1, T2])
        SYS = pnl.System(name='log_test_SYS', processes=[PS])

        def pass_threshold(mech, thresh):
            results = mech.output_states[0].parameters.value.get(SYS)
            for val in results:
                if abs(val) >= thresh:
                    return True
            return False

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.While(pass_threshold, T2, 5.0)
        }

        T1.set_log_conditions(pnl.VALUE)
        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULTS)
        T2.set_log_conditions(pnl.VALUE)
        T2.set_log_conditions('mod_slope')

        SYS.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial)

        log_dict_T1 = T1.log.nparray_dictionary(entries=['RESULTS', 'mod_slope', 'value'])
        log_dict_T2 = T2.log.nparray_dictionary(entries=['value', 'mod_slope'])

        assert list(log_dict_T1.keys()) == [SYS.default_execution_id]
        sys_log_dict = log_dict_T1[SYS.default_execution_id]

        # Check order of keys (must match order of specification)
        assert list(sys_log_dict.keys()) == ['Run', 'Trial', 'Pass', 'Time_step', 'RESULTS', 'mod_slope', 'value']
        assert list(log_dict_T2[SYS.default_execution_id].keys()) == ['Run', 'Trial', 'Pass', 'Time_step', 'value', 'mod_slope']

        # Check values T1
        assert np.allclose(sys_log_dict["Run"], [[0], [0], [0]])
        assert np.allclose(sys_log_dict["Trial"], [[0], [0], [0]])
        assert np.allclose(sys_log_dict["Time_step"], [[0], [0], [0]])
        assert np.allclose(sys_log_dict["RESULTS"], [[0.5], [0.75], [0.875]])
        assert np.allclose(sys_log_dict["value"], [[[0.5]], [[0.75]], [[0.875]]])
        assert np.allclose(sys_log_dict["mod_slope"], [[1], [1], [1]])

        # Check values T2
        assert np.allclose(log_dict_T2[SYS.default_execution_id]["Run"], [[0], [0], [0]])
        assert np.allclose(log_dict_T2[SYS.default_execution_id]["Trial"], [[0], [0], [0]])
        assert np.allclose(log_dict_T2[SYS.default_execution_id]["Time_step"], [[1], [1], [1]])
        assert np.allclose(log_dict_T2[SYS.default_execution_id]["value"], [[[3]], [[4.5]], [[5.25]]])
        assert np.allclose(log_dict_T2[SYS.default_execution_id]["mod_slope"], [[6], [6], [6]])

    def test_log_array_with_scheduler(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.5)
        T2 = pnl.TransferMechanism(name='log_test_T2',
                                   function=psyneulink.core.components.functions.transferfunctions.Linear(slope=6.0))
        PS = pnl.Process(name='log_test_PS', pathway=[T1, T2])
        SYS = pnl.System(name='log_test_SYS', processes=[PS])

        def pass_threshold(mech, thresh):
            results = mech.output_states[0].parameters.value.get(SYS)
            for val in results:
                if abs(val) >= thresh:
                    return True
            return False

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.While(pass_threshold, T2, 5.0)
        }

        T1.set_log_conditions(pnl.VALUE)
        T1.set_log_conditions('mod_slope')
        T1.set_log_conditions(pnl.RESULTS)
        T2.set_log_conditions(pnl.VALUE)
        T2.set_log_conditions('mod_slope')

        SYS.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial)

        log_array_T1 = T1.log.nparray(entries=['RESULTS', 'mod_slope', 'value'])
        log_array_T2 = T2.log.nparray(entries=['value', 'mod_slope'])

        context_results = [pnl.Log.context_header, SYS.default_execution_id]
        # Check values
        run_results = [["Run"], [0], [0], [0]]
        trial_results = [["Trial"], [0], [0], [0]]
        pass_results = [["Pass"], [0], [1], [2]]
        time_step_results = [["Time_step"], [0], [0], [0]]
        results_results = ["RESULTS", [0.5], [0.75], [0.875]]
        slope_results = ["mod_slope", [1], [1], [1]]
        value_results = ["value", [[0.5]], [[0.75]], [[0.875]]]

        for i in range(2):
            assert log_array_T1[0][i] == context_results[i]

        assert log_array_T1[1][0] == pnl.Log.data_header
        data_array = log_array_T1[1][1]
        j = 0
        for i in range(4):
            assert data_array[j][i] == run_results[i]
            assert data_array[j + 1][i] == trial_results[i]
            assert data_array[j + 2][i] == pass_results[i]
            assert data_array[j + 3][i] == time_step_results[i]
            assert data_array[j + 4][i] == results_results[i]
            assert data_array[j + 5][i] == slope_results[i]
            assert data_array[j + 6][i] == value_results[i]

        # Check values
        run_results = [["Run"], [0], [0], [0]]
        trial_results = [["Trial"], [0], [0], [0]]
        pass_results = [["Pass"], [0], [1], [2]]
        time_step_results = [["Time_step"], [1], [1], [1]]
        value_results = ["value", [[3]], [[4.5]], [[5.25]]]
        slope_results = ["mod_slope", [6], [6], [6]]

        for i in range(2):
            assert log_array_T1[0][i] == context_results[i]

        assert log_array_T2[1][0] == pnl.Log.data_header
        data_array = log_array_T2[1][1]
        j = 0
        for i in range(4):
            assert data_array[j][i] == run_results[i]
            assert data_array[j + 1][i] == trial_results[i]
            assert data_array[j + 2][i] == pass_results[i]
            assert data_array[j + 3][i] == time_step_results[i]
            assert data_array[j + 4][i] == value_results[i]
            assert data_array[j + 5][i] == slope_results[i]

    def test_log_dictionary_with_scheduler_many_time_step_increments(self):
        T1 = pnl.TransferMechanism(name='log_test_T1',
                                   integrator_mode=True,
                                   integration_rate=0.05)
        PS = pnl.Process(name='log_test_PS', pathway=[T1])
        SYS = pnl.System(name='log_test_SYS', processes=[PS])

        def pass_threshold(mech, thresh):
            results = mech.output_states[0].parameters.value.get(SYS)
            for val in results:
                if abs(val) >= thresh:
                    return True
            return False

        terminate_trial = {
            pnl.TimeScale.TRIAL: pnl.While(pass_threshold, T1, 0.95)
        }

        T1.set_log_conditions(pnl.VALUE)

        SYS.run(inputs={T1: [[1.0]]}, termination_processing=terminate_trial)

        log_dict_T1 = T1.log.nparray_dictionary(entries=['value'])

        assert list(log_dict_T1.keys()) == [SYS.default_execution_id]
        sys_log_dict = log_dict_T1[SYS.default_execution_id]

        # Check order of keys (must match order of specification)
        assert list(sys_log_dict.keys()) == ['Run', 'Trial', 'Pass', 'Time_step', 'value']

        # # Check values T1
        assert len(sys_log_dict["Run"]) == 59
        assert np.allclose(sys_log_dict["Pass"][30], 30)
        assert np.allclose(sys_log_dict["Time_step"][30], 0)
        assert abs(sys_log_dict["value"][58]) >= 0.95
        assert abs(sys_log_dict["value"][57]) < 0.95

    def test_log_csv_multiple_contexts(self):
        A = pnl.TransferMechanism(name='A')
        B = pnl.TransferMechanism(name='B')
        C = pnl.TransferMechanism(name='C')

        C.set_log_conditions(pnl.VALUE)

        X = pnl.Composition(name='comp X')
        Y = pnl.Composition(name='comp Y')

        X.add_linear_processing_pathway([A, C])
        Y.add_linear_processing_pathway([B, C])

        # running with manual contexts for consistent output
        # because output is sorted by context
        X.run(inputs={A: 1}, context='comp X')
        Y.run(inputs={B: 2}, context='comp Y')

        expected_str = "'Execution Context', 'Data'\n" \
            + "'comp X', 'Run', 'Trial', 'Pass', 'Time_step', 'value'\n" \
            + ", '0', '0', '0', '1', '1.0'\n" \
            + "'comp Y', 'Run', 'Trial', 'Pass', 'Time_step', 'value'\n" \
            + ", '0', '0', '0', '1', '2.0'\n"
        assert C.log.csv() == expected_str

        print()
        print()


class TestClearLog:

    def test_clear_log(self):

        # Create System
        T_1 = pnl.TransferMechanism(name='log_test_T_1', size=2)
        T_2 = pnl.TransferMechanism(name='log_test_T_2', size=2)
        PS = pnl.Process(name='log_test_PS', pathway=[T_1, T_2])
        PJ = T_2.path_afferents[0]
        SYS = pnl.System(name="log_test_SYS", processes=[PS])

        # Set log conditions on each component
        T_1.set_log_conditions('mod_noise')
        T_1.set_log_conditions(pnl.RESULTS)
        T_2.set_log_conditions('mod_slope')
        T_2.set_log_conditions(pnl.RESULTS)
        PJ.set_log_conditions('mod_matrix')

        # Run system
        SYS.run(inputs={T_1: [1.0, 1.0]})

        # Create log dict for each component
        log_dict_T_1 = T_1.log.nparray_dictionary()
        log_dict_T_2 = T_2.log.nparray_dictionary()
        log_dict_PJ = PJ.log.nparray_dictionary()

        assert list(log_dict_T_1.keys()) == [SYS.default_execution_id]
        assert list(log_dict_T_2.keys()) == [SYS.default_execution_id]
        assert list(log_dict_PJ.keys()) == [SYS.default_execution_id]

        # Confirm that values were logged correctly
        sys_log_dict = log_dict_T_1[SYS.default_execution_id]
        assert np.allclose(sys_log_dict['RESULTS'], np.array([[1.0, 1.0]]))
        assert np.allclose(sys_log_dict['mod_noise'], np.array([[0.0]]))

        sys_log_dict = log_dict_T_2[SYS.default_execution_id]
        assert np.allclose(sys_log_dict['RESULTS'], np.array([[1.0, 1.0]]))
        assert np.allclose(sys_log_dict['mod_slope'], np.array([[1.0]]))

        sys_log_dict = log_dict_PJ[SYS.default_execution_id]
        assert np.allclose(sys_log_dict['mod_matrix'], np.array([[1.0, 0.0], [0.0, 1.0]]))

        # KDM 10/3/18: below was changed to delete_entry=True because it's not implemented in Parameter logs,
        # and it's not clear this option results in much difference than just deleting the entries and
        # is stated to be included only for future use
        # Clear T_1s log and DO NOT delete entries
        T_1.log.clear_entries(delete_entry=True)

        # Clear T_2s log and delete entries
        T_2.log.clear_entries(delete_entry=True)

        # Create new log dict for each component
        log_dict_T_1 = T_1.log.nparray_dictionary()
        log_dict_T_2 = T_2.log.nparray_dictionary()
        log_dict_PJ = PJ.log.nparray_dictionary()

        # Confirm that T_1 log values were removed
        assert log_dict_T_1 == OrderedDict()

        # Confirm that T_2 log values were removed and dictionary entries were destroyed
        assert log_dict_T_2 == OrderedDict()

        # Confirm that PJ log values were not affected by changes to T_1 and T_2's logs
        assert np.allclose(log_dict_PJ[SYS.default_execution_id]['mod_matrix'], np.array([[1.0, 0.0], [0.0, 1.0]]))

        # Run system again
        SYS.run(inputs={T_1: [2.0, 2.0]})

        # Create new log dict for each component
        log_dict_T_1 = T_1.log.nparray_dictionary()
        log_dict_T_2 = T_2.log.nparray_dictionary()
        log_dict_PJ = PJ.log.nparray_dictionary()

        # Confirm that T_1 log values only include most recent run
        sys_log_dict = log_dict_T_1[SYS.default_execution_id]
        assert np.allclose(sys_log_dict['RESULTS'], np.array([[2.0, 2.0]]))
        assert np.allclose(sys_log_dict['mod_noise'], np.array([[0.0]]))
        # NOTE: "Run" value still incremented, but only the most recent one is returned (# runs does not reset to zero)
        assert np.allclose(sys_log_dict['Run'], np.array([[1]]))

        # Confirm that T_2 log values only include most recent run
        sys_log_dict = log_dict_T_2[SYS.default_execution_id]
        assert np.allclose(sys_log_dict['RESULTS'], np.array([[2.0, 2.0]]))
        assert np.allclose(sys_log_dict['mod_slope'], np.array([[1.0]]))
        assert np.allclose(sys_log_dict['Run'], np.array([[1]]))

        # Confirm that PJ log values include all runs
        sys_log_dict = log_dict_PJ[SYS.default_execution_id]
        assert np.allclose(sys_log_dict['mod_matrix'], np.array([[[1.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [0.0, 1.0]]]))
        assert np.allclose(sys_log_dict['Run'], np.array([[0], [1]]))

    @pytest.mark.parametrize(
        'insertion_eids, deletion_eids, log_is_empty',
        [
            (['context'], 'context', True),     # fails if string handling not correct due to str being Iterable
            (['context'], ['context'], True),
        ]
    )
    def test_clear_log_arguments(self, insertion_eids, deletion_eids, log_is_empty):
        t = pnl.TransferMechanism()
        c = pnl.Composition()
        c.add_node(t)

        t.parameters.value.log_condition = True

        for eid in insertion_eids:
            c.run({t: 0}, context=eid)

        t.parameters.value.clear_log(deletion_eids)

        if log_is_empty:
            assert len(t.parameters.value.log) == 0
        else:
            assert len(t.parameters.value.log) != 0


class TestFiltering:

    @pytest.fixture(scope='module')
    def node_logged_in_simulation(self):
        Input = pnl.TransferMechanism(name='Input')
        reward = pnl.TransferMechanism(
            output_states=[pnl.RESULT, pnl.OUTPUT_MEAN, pnl.OUTPUT_VARIANCE], name='reward')
        Decision = pnl.DDM(
            function=pnl.DriftDiffusionAnalytical(
                drift_rate=(1.0, pnl.ControlProjection(
                    function=pnl.Linear,
                    control_signal_params={pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)})
                ),
                threshold=(1.0, pnl.ControlProjection(
                    function=pnl.Linear,
                    control_signal_params={pnl.ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)})
                ),
                noise=0.5,
                starting_point=0,
                t0=0.45
            ),
            output_states=[
                pnl.DECISION_VARIABLE,
                pnl.RESPONSE_TIME,
                pnl.PROBABILITY_UPPER_THRESHOLD],
            name='Decision'
        )

        comp = pnl.Composition(name="evc", retain_old_simulation_data=True)
        comp.add_node(reward, required_roles=[pnl.NodeRole.OUTPUT])
        comp.add_node(Decision, required_roles=[pnl.NodeRole.OUTPUT])
        task_execution_pathway = [Input, pnl.IDENTITY_MATRIX, Decision]
        comp.add_linear_processing_pathway(task_execution_pathway)

        comp.add_controller(
            controller=pnl.OptimizationControlMechanism(
                agent_rep=comp,
                features=[Input.input_state, reward.input_state],
                feature_function=pnl.AdaptiveIntegrator(rate=0.5),
                objective_mechanism=pnl.ObjectiveMechanism(
                    function=pnl.LinearCombination(operation=pnl.PRODUCT),
                    monitor=[
                        reward,
                        Decision.output_states[pnl.PROBABILITY_UPPER_THRESHOLD],
                        (Decision.output_states[pnl.RESPONSE_TIME], -1, 1)
                    ]
                ),
                function=pnl.GridSearch(),
                control_signals=[
                    {PROJECTIONS: ("drift_rate", Decision),
                     ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)},
                    {PROJECTIONS: ("threshold", Decision),
                     ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)}
                ]
            )
        )

        comp.enable_controller = True

        stim_list_dict = {
            Input: [0.5, 0.123],
            reward: [20, 20]
        }

        Input.parameters.value.log_condition = True

        comp.run(inputs=stim_list_dict)

        return Input

    def test_node_has_logged_sims(self, node_logged_in_simulation):
        for logged_value, eid_dict in node_logged_in_simulation.log.logged_entries.items():
            for eid in eid_dict:
                if pnl.EID_SIMULATION in str(eid):
                    return
        else:
            assert False, 'No simulation execution_id found in log'

    def test_nparray(self, node_logged_in_simulation):
        for eid in node_logged_in_simulation.log.nparray(exclude_sims=True)[0]:
            assert pnl.EID_SIMULATION not in str(eid)

    def test_nparray_dictionary(self, node_logged_in_simulation):
        for eid in node_logged_in_simulation.log.nparray_dictionary(exclude_sims=True):
            assert pnl.EID_SIMULATION not in str(eid)

    def test_csv(self, node_logged_in_simulation):
        full_csv = node_logged_in_simulation.log.csv(exclude_sims=True)

        # get each row, excluding header
        for row in full_csv.split('\n')[1:]:
            # if present in a row, context will be in the first cell
            assert pnl.EID_SIMULATION not in row.replace("'", '').split(',')[0]



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Scheduler **************************************************************

"""

Overview
--------

A Scheduler is used to generate the order in which the `Components <Component>` of a `Composition <Composition>` are
executed. By default, a Scheduler executes Components in an order determined by the pattern of `Projections
<Projection>` among the `Mechanisms <Mechanism>` in the `Composition <Composition>`, with each Mechanism executed once
per `PASS` through the Composition. For example, in a `System` in which a Mechanism A projects to a Mechanism B that
projects to a Mechanism C, A will execute first followed by B, and then C in each `PASS` through the System.  However,
a Scheduler can be used to implement more complex patterns of execution, by specifying `Conditions <Condition>` that
determine when and how many times individual Components execute, and whether and how this depends on the execution of
other Components. Any executable Component in a Composition can be assigned a Condition, and Conditions can be combined
in arbitrary ways to generate any pattern of execution of the Components in a Composition that is logically possible.

.. note::
   In general, `Mechanisms <Mechanism>` are the Components of a Composition that are most commonly associated with
   Conditions, and assigned to a Scheduler for execution.  However, in some circumstances, `Projections <Projection>`
   can also be assigned for execution (e.g., during `learning <Process_Execution_Learning>` to insure that
   `MappingProjections <MappingProjection>` are updated in the proper order).

.. _Scheduler_Creation:

Creating a Scheduler
--------------------

A Scheduler can be created explicitly using its constructor.  However, more commonly it is created automatically
for a `Composition <Composition>` when it is created.  When creating a Scheduler explicitly, the set of `Components
<Component>` to be executed and their order must be specified in the Scheduler's constructor using one the following:

COMMENT:
   JDC: WE MAY WANT TO CHANGE THE NAME OF THE ARGUMENT TO 'COMPOSITION` ONCE THAT IS IMPLEMENTED, TO BE FULLY GENERAL
COMMENT

* a `Composition` in the **composition** argument - if a Composition is specified,
  the Scheduler is created using the nodes and edges in the Composition's `graph <Composition.graph_processing>`.

* a *graph specification dictionary* in the **graph** argument -
  each entry of the dictionary must be a Component of a Composition, and the value of each entry must be a set of
  zero or more Components that project directly to the key.  The graph must be acyclic; an error is generated if any
  cycles (e.g., recurrent dependencies) are detected.  The Scheduler computes a `toposort` from the graph that is
  used as the default order of executions, subject to any `Condition`s that have been specified
  (see `below <Scheduler_Algorithm>`).

If both a Composition and a graph are specified, the Composition takes precedence, and the graph is ignored.

Conditions can be added to a Scheduler when it is created by specifying a `ConditionSet` (a set of
`Conditions <Condition>`) in the **conditions** argument of its constructor.  Individual Conditions and/or
ConditionSets can also be added after the  Scheduler has been created, using its `add_condition` and
`add_condition_set` methods, respectively.

.. _Scheduler_Algorithm:

Algorithm
---------

When a Scheduler is created, it constructs a `consideration_queue`:  a list of `consideration_sets <consideration_set>`
that defines the order in which Components are eligible to be executed.  This is based on the pattern of projections
among them specified in the Composition, or on the dependencies specified in the graph specification dictionary, whichever
was provided in the Scheduler's constructor.  Each `consideration_set` is a set of Components that are eligible to
execute at the same time/`TIME_STEP` (i.e., that appear at the same "depth" in a sequence of dependencies, and
among which there are no dependencies).  The first `consideration_set` consists of only `ORIGIN` Mechanisms.
The second consists of all Components that receive `Projections <Projection>` from the Mechanisms in the first
`consideration_set`. The third consists of Components that receive Projections from Components in the first two
`consideration_sets <consideration_set>`, and so forth.  When the Scheduler is run, it uses the
`consideration_queue` to determine which Components are eligible to execute in each `TIME_STEP` of a `PASS`, and then
evaluates the `Condition <Condition>` associated with each Component in the current `consideration_set`
to determine which should actually be assigned for execution.

Pseudocode::

    consideration_queue <- list(toposort(graph))

    reset TimeScale.TRIAL counters
    while TimeScale.TRIAL termination conditions are not satisfied:
        reset TimeScale.PASS counters
        cur_index <- 0

        while TimeScale.TRIAL termination conditions are not satisfied
              and cur_index < len(consideration_queue):

            cur_consideration_set <- consideration_queue[cur_index]
            do:
                cur_consideration_set_has_changed <- False
                for cur_node in cur_consideration_set:
                    if  cur_node not in cur_time_step
                        and cur_node`s Condition is satisfied:

                        cur_consideration_set_has_changed <- True
                        add cur_node to cur_time_step
                        increment execution and time counters
            while cur_consideration_set_has_changed

            if cur_time_step is not empty:
                yield cur_time_step

            increment cur_index
            increment time counters

.. _Scheduler_Execution:

Execution
---------

When a Scheduler is run, it provides a set of Components that should be run next, based on their dependencies in the
Composition or graph specification dictionary, and any `Conditions <Condition>`, specified in the Scheduler's constructor.
For each call to the `run <Scheduler.run>` method, the Scheduler sequentially evaluates its
`consideration_sets <consideration_set>` in their order in the `consideration_queue`.  For each set, it  determines
which Components in the set are allowed to execute, based on whether their associated `Condition <Condition>` has
been met. Any Component that does not have a `Condition` explicitly specified is assigned a Condition that causes it
to be executed whenever it is `under consideration <Scheduler_Algorithm>` and all its structural parents have been executed
at least once since the Component's last execution. All of the Components within a `consideration_set` that
are allowed to execute comprise a `TIME_STEP` of execution. These Components are
considered as executing simultaneously.

.. note::
    The ordering of the Components specified within a `TIME_STEP` is arbitrary
    (and is irrelevant, as there are no graph dependencies among Components within the same `consideration_set`).
    However, the execution of a Component within a `time_step` may trigger the execution of another Component within its
    `consideration_set`, as in the example below::

            C
          ↗ ↖
         A   B

        scheduler.add_condition(B, pnl.EveryNCalls(A, 2))
        scheduler.add_condition(C, pnl.EveryNCalls(B, 1))

        time steps: [{A}, {A, B}, {C}, ...]

    Since there are no graph dependencies between `A` and `B`, they may execute in the same `TIME_STEP`. Morever,
    `A` and `B` are in the same `consideration_set`. Since `B` is specified to run every two times `A` runs,
    `A`'s second execution in the second `TIME_STEP` allows `B` to run within that `TIME_STEP`, rather
    than waiting for the next `PASS`.

For each `TIME_STEP`, the Scheduler evaluates  whether any specified
`termination Conditions <Scheduler_Termination_Conditions>` have been met, and terminates if so.  Otherwise,
it returns the set of Components that should be executed in the current `TIME_STEP`. Each subsequent call to the
`run <Scheduler.run>` method returns the set of Components in the following `TIME_STEP`.

Processing of all of the `consideration_sets <consideration_set>` in the `consideration_queue` constitutes a `PASS` of
execution, over which every Component in the Composition has been considered for execution. Subsequent calls to the
`run <Scheduler.run>` method cycle back through the `consideration_queue`, evaluating the
`consideration_sets <consideration_set>` in the same order as previously. Different subsets of Components within the
same `consideration_set` may be assigned to execute on each `PASS`, since different Conditions may be satisfied.

The Scheduler continues to make `PASS`es through the `consideration_queue` until a
`termination Condition <Scheduler_Termination_Conditions>` is satisfied. If no termination Conditions are specified,
the Scheduler terminates a `TRIAL` when every Component has been specified for execution at least once (corresponding
to the `AllHaveRun` Condition).  However, other termination Conditions can be specified, that may cause the Scheduler
to terminate a `TRIAL` earlier  or later (e.g., when the  Condition for a particular Component or set of Components
is met).  When the Scheduler terminates a `TRIAL`, the `Composition <Composition>` begins processing the next input
specified in the call to its `run <Composition.run>` method.  Thus, a `TRIAL` is defined as the scope of processing
associated with a given input to the Composition.


.. _Scheduler_Termination_Conditions:

*Termination Conditions*
~~~~~~~~~~~~~~~~~~~~~~~~

Termination conditions are `Conditions <Condition>` that specify when the open-ended units of time - `TRIAL`
and `RUN` - have ended.  By default, the termination condition for a `TRIAL` is `AllHaveRun`, which is satisfied
when all Components have run at least once within the trial, and the termination condition for a `RUN` is
when all of its constituent trials have terminated. These defaults may be overriden when running a Composition,
by passing a dictionary mapping `TimeScales <TimeScale>` to `Conditions <Condition>` in the
**termination_processing** argument of a call to `Composition.run` (to terminate the execution of processing)::

    Composition.run(
        ...,
        termination_processing={TimeScale.TRIAL: WhenFinished(ddm)}
        )

Examples
--------

Please see `Condition` for a list of all supported Conditions and their behavior.

* Basic phasing in a linear process::

    >>> import psyneulink as pnl

    >>> A = pnl.TransferMechanism(name='A')
    >>> B = pnl.TransferMechanism(name='B')
    >>> C = pnl.TransferMechanism(name='C')

    >>> comp = pnl.Composition()
    >>> comp.add_linear_processing_pathway([A, B, C])
    [(TransferMechanism A), (MappingProjection MappingProjection from A[RESULTS] to B[InputState-0]), (TransferMechanism B), (MappingProjection MappingProjection from B[RESULTS] to C[InputState-0]), (TransferMechanism C)]

    >>> # implicit condition of Always for A
    >>> comp.scheduler_processing.add_condition(B, pnl.EveryNCalls(A, 2))
    >>> comp.scheduler_processing.add_condition(C, pnl.EveryNCalls(B, 3))

    >>> # implicit AllHaveRun Termination condition
    >>> execution_sequence = list(comp.scheduler_processing.run())
    >>> execution_sequence
    [{(TransferMechanism A)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism C)}]

* Alternate basic phasing in a linear process::

    >>> comp = pnl.Composition()
    >>> comp.add_linear_processing_pathway([A, B])
    [(TransferMechanism A), (MappingProjection MappingProjection-2), (TransferMechanism B)]

    >>> comp.scheduler_processing.add_condition(
    ...     A,
    ...     pnl.Any(
    ...         pnl.AtPass(0),
    ...         pnl.EveryNCalls(B, 2)
    ...     )
    ... )

    >>> comp.scheduler_processing.add_condition(
    ...     B,
    ...     pnl.Any(
    ...         pnl.EveryNCalls(A, 1),
    ...         pnl.EveryNCalls(B, 1)
    ...     )
    ... )
    >>> termination_conds = {
    ...     pnl.TimeScale.TRIAL: pnl.AfterNCalls(B, 4, time_scale=pnl.TimeScale.TRIAL)
    ... }
    >>> execution_sequence = list(comp.scheduler_processing.run(termination_conds=termination_conds))
    >>> execution_sequence # doctest: +SKIP
    [{(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism B)}, {(TransferMechanism B)}]

* Basic phasing in two processes::

    >>> comp = pnl.Composition()
    >>> comp.add_linear_processing_pathway([A, C])
    [(TransferMechanism A), (MappingProjection MappingProjection from A[RESULTS] to C[InputState-0]), (TransferMechanism C)]

    >>> comp.add_linear_processing_pathway([B, C])
    [(TransferMechanism B), (MappingProjection MappingProjection-4), (TransferMechanism C)]

    >>> comp.scheduler_processing.add_condition(A, pnl.EveryNPasses(1))
    >>> comp.scheduler_processing.add_condition(B, pnl.EveryNCalls(A, 2))
    >>> comp.scheduler_processing.add_condition(
    ...     C,
    ...     pnl.Any(
    ...         pnl.AfterNCalls(A, 3),
    ...         pnl.AfterNCalls(B, 3)
    ...     )
    ... )
    >>> termination_conds = {
    ...     pnl.TimeScale.TRIAL: pnl.AfterNCalls(C, 4, time_scale=pnl.TimeScale.TRIAL)
    ... }
    >>> execution_sequence = list(comp.scheduler_processing.run(termination_conds=termination_conds))
    >>> execution_sequence  # doctest: +SKIP
    [{(TransferMechanism A)}, {(TransferMechanism A), (TransferMechanism B)}, {(TransferMechanism A)}, {(TransferMechanism C)}, {(TransferMechanism A), (TransferMechanism B)}, {(TransferMechanism C)}, {(TransferMechanism A)}, {(TransferMechanism C)}, {(TransferMechanism A), (TransferMechanism B)}, {(TransferMechanism C)}]

.. _Scheduler_Class_Reference

Class Reference
===============

"""

import collections
import copy
import datetime
import logging
import warnings

from toposort import toposort, toposort_flatten

from psyneulink.core.globals.context import Context, handle_external_context
from psyneulink.core.scheduling.condition import All, AllHaveRun, Always, Condition, ConditionSet, EveryNCalls, Never
from psyneulink.core.scheduling.time import Clock, TimeScale

__all__ = [
    'Scheduler', 'SchedulerError',
]

logger = logging.getLogger(__name__)


class SchedulerError(Exception):

    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class Scheduler(object):
    """Generates an order of execution for `Components <Component>` in a `Composition <Composition>` or graph
    specification dictionary, possibly determined by a set of `Conditions <Condition>`.

    Arguments
    ---------

    composition : Composition
        specifies the Components to be ordered for execution, and any dependencies among them, based on the
        Composition's `graph <Composition.graph_processing>`.

    COMMENT:
        [**??IS THE FOLLOWING CORRECT]:
        K: not correct, there are no implicit System Conditions
        JDC: I WAS REFERRING TO THE DEPENDENCIES IN THE SYSTEM'S GRAPH.  THE FACT THAT conditions IS AN
             OPTIONAL ARG FOR SCHEDULER, AND THAT PROVIDING A System IS SUFFICIENT TO GENERATE A SCHEDULE,
             MEANS THAT THERE MUST BE CONDITIONS IMPLICIT IN THE System.
        K: it's not that they're implicit, it's that we just set defaults to match the behavior of the
            naive scheduler
    COMMENT

    conditions  : ConditionSet
        set of `Conditions <Condition>` that specify when individual Components in **composition**
        execute and any dependencies among them

    graph : Dict[Component: set(Component)]
        a graph specification dictionary - each entry of the dictionary must be a Component,
        and the value of each entry must be a set of zero or more Components that project directly to the key.

    Attributes
    ----------

    conditions : ConditionSet
        the set of Conditions the Scheduler uses when running

    execution_list : list
        the full history of time steps the Scheduler has produced

    consideration_queue: list
        a list form of the Scheduler's toposort ordering of its nodes

    termination_conds : Dict[TimeScale: Condition]
        a mapping from `TimeScales <TimeScale>` to `Conditions <Condition>` that, when met, terminate the execution
        of the specified `TimeScale`.

    times: Dict[TimeScale: Dict[TimeScale: int]]
        a structure counting the number of occurrences of a certain `TimeScale` within the scope of another `TimeScale`.
        For example, `times[TimeScale.RUN][TimeScale.PASS]` is the number of `PASS`es that have occurred in the
        current `RUN` that the Scheduler is scheduling at the time it is accessed

    clock : `Clock`
        a `Clock` object that stores the current time in this Scheduler

    """
    def __init__(
        self,
        composition=None,
        graph=None,
        conditions=None,
        termination_conds={
            TimeScale.RUN: Never(),
            TimeScale.TRIAL: AllHaveRun(),
        },
        default_execution_id=None,
        **kwargs
    ):
        """
        :param self:
        :param composition: (Composition) - the Composition this scheduler is scheduling for
        :param conditions: (ConditionSet) - a :keyword:`ConditionSet` to be scheduled
        """
        self.conditions = conditions if conditions is not None else ConditionSet()

        # stores the in order list of self.run's yielded outputs
        self.consideration_queue = []
        self.default_termination_conds = Scheduler._parse_termination_conditions(termination_conds)
        self._termination_conds = termination_conds.copy()

        self.cycle_nodes = set()

        # can remove this once system is fully eliminated from tests/support
        try:
            system = kwargs['system']
            warnings.warn('Systems are no longer updated, new code should use Compositions.', DeprecationWarning)
        except KeyError:
            system = None

        if composition is not None:
            self.nodes = [vert.component for vert in composition.graph_processing.vertices]
            self._init_consideration_queue_from_graph(composition.graph_processing)
            if default_execution_id is None:
                default_execution_id = composition.default_execution_id
        elif system is not None:
            self.nodes = [m for m in system.execution_list]
            self._init_consideration_queue_from_system(system)
            if default_execution_id is None:
                default_execution_id = system.default_execution_id
        elif graph is not None:
            try:
                self.nodes = [vert.component for vert in graph.vertices]
                self._init_consideration_queue_from_graph(graph)
            except AttributeError:
                self.consideration_queue = list(toposort(graph))
                self.nodes = []
                for consideration_set in self.consideration_queue:
                    for node in consideration_set:
                        self.nodes.append(node)
        else:
            raise SchedulerError('Must instantiate a Scheduler with either a Composition (kwarg composition) '
                                 'or a graph dependency dict (kwarg graph)')

        self.default_execution_id = default_execution_id
        self.execution_list = {self.default_execution_id: []}
        self.clocks = {self.default_execution_id: Clock()}
        self.counts_total = {}
        self.counts_useable = {}
        self._init_counts(execution_id=self.default_execution_id)
        self.date_creation = datetime.datetime.now()
        self.date_last_run_end = None

    # the consideration queue is the ordered list of sets of nodes in the graph, by the
    # order in which they should be checked to ensure that all parents have a chance to run before their children
    def _init_consideration_queue_from_system(self, system):
        dependencies = []
        for dependency_set in list(toposort(system.execution_graph)):
            new_set = set()
            for d in dependency_set:
                new_set.add(d)
            dependencies.append(new_set)
        self.consideration_queue = dependencies

    def _dfs_for_cycles(self, dependencies, node, loop_start_set, visited, loop):

        if node in loop_start_set:
            loop.append(node)
            return loop

        if visited is None:
            visited = set()
        visited.add(node)
        loop.append(node)

        if len(loop) == 2:
            for next_node in dependencies[node]:
                if next_node in loop_start_set:
                    loop.append(next_node)
                    return loop

        for next_node in dependencies[node] - visited:
            return self._dfs_for_cycles(dependencies, next_node, loop_start_set, visited, loop)

    def _call_toposort(self, graph):
        """
        execution_depenencies stored in self.depdency_sets
        :return:
        """

        execution_dependencies = {}         # stores  a modified version of the graph in which cycles are "flattened"
        removed_dependencies = {}           # stores dependencies that were removed in order to flatten cycles
        flattened_cycles = {}               # flattened_cycles[node] = [all cycles to which node belongs]
        structural_dependencies = collections.OrderedDict()
        self.cycle_nodes = set()
        # Loop through the existing composition graph, considering "forward" projections only
        # If a cycle is found, "flatten" it by bringing all nodes into the same execution set
        for vert in graph.vertices:

            if vert.component not in execution_dependencies:
                execution_dependencies[vert.component] = set()
                structural_dependencies[vert.component] = set()

            # use "get_forward_children_from_component" to ignore any projections that were marked as "feedback"
            # "feedback" projections, we've already determined, happen after all forward projections, but when "forward"
            # projections cause cycles, we need to execute them all at once
            for child in graph.get_forward_children_from_component(vert.component):
                if child.component not in execution_dependencies:
                    execution_dependencies[child.component] = set()
                    structural_dependencies[child.component] = set()
                execution_dependencies[child.component].add(vert.component)

                # loop_start_set contains the current starting point and any cycles it is already connected to
                # if the new dependency introduces any paths that lead back to a node in loop_start_set, then
                # we will consider the new path a cycle
                loop_start_set = {child.component}
                connected_cycles = set()
                self._get_all_connected_cycles(connected_cycles, child.component, set(), flattened_cycles)
                for node in connected_cycles:
                    loop_start_set.add(node)

                # if the new dependency created a cycle, return that cycle
                cycle = self._dfs_for_cycles(execution_dependencies, vert.component, loop_start_set, None, [child.component])

                if cycle:
                    # loop over all nodes in the cycle in order to:
                    # (1) store the node: cycle pair in the flattened cycles dict
                    # (2) remove the dependencies that created the cycle
                    # (3) copy the dependencies of the node that "started" the cycle onto all other cycle nodes
                    for i in range(len(cycle) - 1):
                        node_a = cycle[i]
                        self.cycle_nodes.add(node_a)
                        node_b = cycle[i + 1]
                        if node_a not in flattened_cycles:
                            flattened_cycles[node_a] = []
                        flattened_cycles[node_a].append(cycle)
                        execution_dependencies[node_a].remove(node_b)
                        if node_a not in removed_dependencies:
                            removed_dependencies[node_a] = set()
                        removed_dependencies[node_a].add(node_b)

                        if i != 0:
                            for dependency in execution_dependencies[cycle[0]]:
                                execution_dependencies[cycle[i]].add(dependency)
                else:
                    # necessary for the case where you want to add a projection that terminates at a node in a loop
                    # AFTER the loop has already been created.
                    # e.g. ORIGINAL:    A <--> B <--> C -- > D
                    # NEW: new_node --> A <--> B <--> C -- > D
                    # (otherwise, the order in which a user adds components to a composition would affect the graph)
                    for cycle_node in connected_cycles:
                        execution_dependencies[cycle_node].add(vert.component)
                structural_dependencies[child.component].add(vert.component)
            for child in graph.get_backward_children_from_component(vert.component):
                if child.component not in execution_dependencies:
                    structural_dependencies[child.component] = set()
                structural_dependencies[child.component].add(vert.component)
            self.dependency_dict = execution_dependencies

        return list(toposort(execution_dependencies)), removed_dependencies, structural_dependencies

    def _get_all_connected_cycles(self, connected_cycles, original_key, visited_keys, flattened_cycles):
        if original_key in flattened_cycles:
            if original_key in visited_keys:
                return
            cycles = flattened_cycles[original_key]
            visited_keys.add(original_key)
            for cycle in cycles:
                for cycle_node in cycle:
                    connected_cycles.add(cycle_node)
                    self._get_all_connected_cycles(connected_cycles, cycle_node, visited_keys, flattened_cycles)

    def _init_consideration_queue_from_graph(self, graph):
        self.consideration_queue, self.removed_dependencies, self.structural_dependencies = self._call_toposort(graph)

    def _init_counts(self, execution_id=None, base_execution_id=None):
        """
            Attributes
            ----------

                execution_id
                    the execution_id to initialize counts for
                    default : self.default_execution_id

                base_execution_id
                    if specified, the counts for execution_id will be copied from the counts of base_execution_id
                    default : None
        """
        # all counts are divided by execution_id, which provides a context for the scheduler's execution, so that
        # it can be reused in multiple contexts

        # stores total the number of occurrences of a node through the time scale
        # i.e. the number of times node has ran/been queued to run in a trial
        if execution_id not in self.counts_total:
            self.counts_total[execution_id] = {}

            if base_execution_id is not None:
                if base_execution_id not in self.counts_total:
                    raise SchedulerError('execution_id {0} not in {1}.counts_total'.format(base_execution_id, self))

                self.counts_total[execution_id] = {
                    ts: {n: self.counts_total[base_execution_id][ts][n] for n in self.nodes} for ts in TimeScale
                }
            else:
                self.counts_total[execution_id] = {
                    ts: {n: 0 for n in self.nodes} for ts in TimeScale
                }

        # counts_useable is a dictionary intended to store the number of available "instances" of a certain node that
        # are available to expend in order to satisfy conditions such as "run B every two times A runs"
        # specifically, counts_useable[a][b] = n indicates that there are n uses of a that are available for b to expend
        # so, in the previous example B would check to see if counts_useable[A][B] >= 2, in which case B can run
        # then, counts_useable[a][b] would be reset to 0, even if it was greater than 2
        if execution_id not in self.counts_useable:
            self.counts_useable[execution_id] = {}

            if base_execution_id is not None:
                if base_execution_id not in self.counts_useable:
                    raise SchedulerError('execution_id {0} not in {1}.counts_useable'.format(base_execution_id, self))

                self.counts_useable[execution_id] = {
                    node: {n: self.counts_useable[base_execution_id][node][n] for n in self.nodes} for node in self.nodes
                }
            else:
                self.counts_useable[execution_id] = {
                    node: {n: 0 for n in self.nodes} for node in self.nodes
                }

        if execution_id not in self.execution_list:
            if base_execution_id is not None:
                if base_execution_id not in self.execution_list:
                    raise SchedulerError('execution_id {0} not in {1}.execution_list'.format(base_execution_id, self))

                self.execution_list[execution_id] = list(self.execution_list[base_execution_id])
            else:
                self.execution_list[execution_id] = []

        self._init_clock(execution_id, base_execution_id)

    def _init_clock(self, execution_id=None, base_execution_id=None):
        # instantiate new Clock for this execution_id if necessary
        # currently does not work with base_execution_id
        if execution_id not in self.clocks:
            if base_execution_id is not None:
                if base_execution_id not in self.clocks:
                    raise SchedulerError('execution_id {0} not in {1}.clocks'.format(base_execution_id, self))

                self.clocks[execution_id] = copy.deepcopy(self.clocks[base_execution_id])
            else:
                self.clocks[execution_id] = Clock()

    def _reset_counts_total(self, time_scale, execution_id=None):
        for ts in TimeScale:
            # only reset the values underneath the current scope
            # this works because the enum is set so that higher granularities of time have lower values
            if ts.value <= time_scale.value:
                for c in self.counts_total[execution_id][ts]:
                    self.counts_total[execution_id][ts][c] = 0

    def _reset_counts_useable(self, execution_id=None):
        self.counts_useable[execution_id] = {
            node: {n: 0 for n in self.nodes} for node in self.nodes
        }

    def update_termination_conditions(self, termination_conds):
        termination_conds = Scheduler._parse_termination_conditions(termination_conds)
        new_conds = self.termination_conds.copy()
        new_conds.update(termination_conds)

        return new_conds

    @staticmethod
    def _parse_termination_conditions(termination_conds):
        try:
            return {k: termination_conds[k] for k in termination_conds if isinstance(k, TimeScale) and isinstance(termination_conds[k], Condition)}
        except TypeError:
            raise TypeError('termination_conditions must be a dictionary of the form {TimeScale: Condition, ...}')

    ################################################################################
    # Wrapper methods
    #   to allow the user to ignore the ConditionSet internals
    ################################################################################
    def __contains__(self, item):
        return self.conditions.__contains__(item)

    def add_condition(self, owner, condition):
        """
        Adds a `Condition` to the Scheduler. If **owner** already has a Condition, it is overwritten
        with the new one. If you want to add multiple conditions to a single owner, use the
        `composite Conditions <Conditions_Composite>` to accurately specify the desired behavior.

        Arguments
        ---------

        owner : Component
            specifies the Component with which the **condition** should be associated. **condition**
            will govern the execution behavior of **owner**

        condition : Condition
            specifies the Condition, associated with the **owner** to be added to the ConditionSet.
        """
        self.conditions.add_condition(owner, condition)

    def add_condition_set(self, conditions):
        """
        Adds a set of `Conditions <Condition>` (in the form of a dict or another ConditionSet) to the Scheduler.
        Any Condition added here will overwrite an existing Condition for a given owner.
        If you want to add multiple conditions to a single owner, add a single `Composite Condition <Conditions_Composite>`
        to accurately specify the desired behavior.

        Arguments
        ---------

        conditions : dict[`Component <Component>`: `Condition`], `ConditionSet`
            specifies collection of Conditions to be added to this ConditionSet,

            if a dict is provided:
                each entry should map an owner `Component` (the `Component` whose execution behavior will be
                governed) to a `Condition <Condition>`

        """
        self.conditions.add_condition_set(conditions)

    ################################################################################
    # Validation methods
    #   to provide the user with info if they do something odd
    ################################################################################
    def _validate_run_state(self):
        self._validate_conditions()

    def _validate_conditions(self):
        unspecified_nodes = []
        for node in self.nodes:
            if node not in self.conditions:
                # determine parent nodes
                node_index = 0
                for i in range(len(self.consideration_queue)):
                    if node in self.consideration_queue[i]:
                        node_index = i
                        break

                if node_index > 0:
                    dependencies = list(self.consideration_queue[i - 1])
                    if len(dependencies) == 1:
                        cond = EveryNCalls(dependencies[0], 1)
                    elif len(dependencies) > 1:
                        cond = All(*[EveryNCalls(x, 1) for x in dependencies])
                    else:
                        raise SchedulerError(f'{self}: Empty consideration set in consideration_queue[{i - 1}]')
                else:
                    cond = Always()

                self.conditions.add_condition(node, cond)
                unspecified_nodes.append(node)
        if len(unspecified_nodes) > 0:
            logger.info(
                'These nodes have no Conditions specified, and will be scheduled with conditions: {0}'.format(
                    {node: self.conditions[node] for node in unspecified_nodes}
                )
            )

    ################################################################################
    # Run methods
    ################################################################################

    def run(self, termination_conds=None, context=None, base_context=Context(execution_id=None), skip_trial_time_increment=False):
        """
        run is a python generator, that when iterated over provides the next `TIME_STEP` of
        executions at each iteration

        :param termination_conds: (dict) - a mapping from `TimeScale`\\s to `Condition`\\s that when met
               terminate the execution of the specified `TimeScale`
        """
        self._validate_run_state()
        if termination_conds is None:
            termination_conds = self.termination_conds
        else:
            termination_conds = self.update_termination_conditions(Scheduler._parse_termination_conditions(termination_conds))

        if context is None:
            context = Context(execution_id=self.default_execution_id)

        self._init_counts(context.execution_id, base_context.execution_id)
        self._reset_counts_useable(context.execution_id)
        self._reset_counts_total(TimeScale.TRIAL, context.execution_id)

        while (
            not termination_conds[TimeScale.TRIAL].is_satisfied(scheduler=self, context=context)
            and not termination_conds[TimeScale.RUN].is_satisfied(scheduler=self, context=context)
        ):
            self._reset_counts_total(TimeScale.PASS, context.execution_id)

            execution_list_has_changed = False
            cur_index_consideration_queue = 0

            while (
                cur_index_consideration_queue < len(self.consideration_queue)
                and not termination_conds[TimeScale.TRIAL].is_satisfied(scheduler=self, context=context)
                and not termination_conds[TimeScale.RUN].is_satisfied(scheduler=self, context=context)
            ):
                # all nodes to be added during this time step
                cur_time_step_exec = set()
                # the current "layer/group" of nodes that MIGHT be added during this time step
                cur_consideration_set = self.consideration_queue[cur_index_consideration_queue]
                try:
                    iter(cur_consideration_set)
                except TypeError as e:
                    raise SchedulerError('cur_consideration_set is not iterable, did you ensure that this Scheduler was instantiated with an actual toposort output for param toposort_ordering? err: {0}'.format(e))

                # do-while, on cur_consideration_set_has_changed
                # we check whether each node in the current consideration set is allowed to run,
                # and nodes can cause cascading adds within this set
                while True:
                    cur_consideration_set_has_changed = False
                    for current_node in cur_consideration_set:
                        # only add each node once during a single time step, this also serves
                        # to prevent infinitely cascading adds
                        if current_node not in cur_time_step_exec:
                            if self.conditions.conditions[current_node].is_satisfied(scheduler=self, context=context):
                                cur_time_step_exec.add(current_node)
                                execution_list_has_changed = True
                                cur_consideration_set_has_changed = True

                                for ts in TimeScale:
                                    self.counts_total[context.execution_id][ts][current_node] += 1
                                # current_node's node is added to the execution queue, so we now need to
                                # reset all of the counts useable by current_node's node to 0
                                for n in self.counts_useable[context.execution_id]:
                                    self.counts_useable[context.execution_id][n][current_node] = 0
                                # and increment all of the counts of current_node's node useable by other
                                # nodes by 1
                                for n in self.counts_useable[context.execution_id]:
                                    self.counts_useable[context.execution_id][current_node][n] += 1
                    # do-while condition
                    if not cur_consideration_set_has_changed:
                        break

                # add a new time step at each step in a pass, if the time step would not be empty
                if len(cur_time_step_exec) >= 1:
                    self.execution_list[context.execution_id].append(cur_time_step_exec)
                    yield self.execution_list[context.execution_id][-1]

                    self.get_clock(context)._increment_time(TimeScale.TIME_STEP)

                cur_index_consideration_queue += 1

            # if an entire pass occurs with nothing running, add an empty time step
            if not execution_list_has_changed:
                self.execution_list[context.execution_id].append(set())
                yield self.execution_list[context.execution_id][-1]

                self.get_clock(context)._increment_time(TimeScale.TIME_STEP)

            self.get_clock(context)._increment_time(TimeScale.PASS)

        if not skip_trial_time_increment:
            self.get_clock(context)._increment_time(TimeScale.TRIAL)

        if termination_conds[TimeScale.RUN].is_satisfied(scheduler=self, context=context):
            self.date_last_run_end = datetime.datetime.now()

        return self.execution_list[context.execution_id]

    def _dict_summary(self):
        return {
            'conditions': {
                'termination': {
                    str(k): str(v) for k, v in self.termination_conds.items()
                },
                'node': {
                    n.name: str(self.conditions[n]) for n in self.nodes if n in self.conditions
                }
            }
        }

    def json_summary(self):
        import json
        return json.dumps(self._dict_summary(), sort_keys=True, indent=4, separators=(',', ': '))

    @property
    def clock(self):
        return self.clocks[self.default_execution_id]

    def get_clock(self, context):
        try:
            return self.clocks[context.default_execution_id]
        except AttributeError:
            try:
                return self.clocks[context.execution_id]
            except AttributeError:
                return self.clocks[context]
        except KeyError:
            raise

    @property
    def termination_conds(self):
        return self._termination_conds

    @termination_conds.setter
    def termination_conds(self, termination_conds):
        if termination_conds is None:
            self._termination_conds = self.default_termination_conds.copy()
        else:
            self._termination_conds.update(termination_conds)



import functools
import numpy as np
import psyneulink as pnl

# Seed random number generators for consistency in testing
# seed = 0
# random.seed(seed)
# np.random.seed(seed)

# CONSTRUCTION:
import psyneulink.core.components.functions.distributionfunctions
import psyneulink.core.components.functions.learningfunctions

input_layer = pnl.TransferMechanism(
    size=2,
    name='Input Layer'
)

# Takes sum of input layer elements as external component of drift rate
# Notes:
#    - drift_rate parameter in constructor for DDM is the "internally modulated" component of the drift_rate;
#    - arguments to DDM's function (DriftDiffusionAnalytical) are specified as CONTROL, so that their values will be determined
#        by the EVCControlMechanism of the System to which the action_selection Mechanism is assigned (see below)
#    - the input_format argument specifies that the input to the DDM should be one-hot encoded two-element array
#    - the output_states argument specifies use of the DECISION_VARIABLE_ARRAY OutputState, which encodes the
#        response in the same format as the ARRAY input_format/.
action_selection = pnl.DDM(
        input_format=pnl.ARRAY,
        function=psyneulink.core.components.functions.distributionfunctions.DriftDiffusionAnalytical(
                drift_rate=pnl.CONTROL,
                threshold=pnl.CONTROL,
                starting_point=pnl.CONTROL,
                noise=pnl.CONTROL,
        ),
        output_states=[pnl.SELECTED_INPUT_ARRAY],
        name='DDM'
)

# Construct Process
# Notes:
#    The np.array specifies the matrix used as the Mapping Projection from input_layer to action_selection,
#        which insures the left element of the input favors the left action (positive value of DDM decision variable),
#        and the right element favors the right action (negative value of DDM decision variable)
#    The learning argument specifies Reinforcement as the learning function for the Projection
p = pnl.Process(
    default_variable=[0, 0],
    # pathway=[input_layer, np.array([[1],[-1]]), action_selection],
    pathway=[input_layer, pnl.IDENTITY_MATRIX, action_selection],
    learning=pnl.LearningProjection(learning_function=psyneulink.core.components.functions.learningfunctions
                                    .Reinforcement(learning_rate=0.5)),
    target=0
)

s = pnl.System(
    processes=[p],
    controller=pnl.EVCControlMechanism(control_signals=(pnl.LEARNING_RATE, p.learning_mechanisms[1]))
)

# EXECUTION:

# Prints initial weight matrix for the Projection from the input_layer to the action_selection Mechanism
print('reward prediction weights: \n', action_selection.input_state.path_afferents[0].matrix)


# Used by *call_before_trial* and *call_after_trial* to generate printouts.
# Note:  should be replaced by use of logging functionality that has now been implemented.
def print_header(system):
    print("\n\n**** Time: ", system.scheduler_processing.get_clock(system).simple_time)


def show_weights(context=None):
    print('\nReward prediction weights: \n', action_selection.input_state.path_afferents[0].get_mod_matrix(context))
    # print(
    #     '\nAction selected:  {}; predicted reward: {}'.format(
    #         np.nonzero(action_selection.output_state.value)[0][0],
    #         action_selection.output_state.value[np.nonzero(action_selection.output_state.value)][0]
    #     )
    # )
    comparator = action_selection.output_state.efferents[0].receiver.owner
    learn_mech = action_selection.output_state.efferents[1].receiver.owner
    print('\nact_sel_in_state variable:  {} '
          '\nact_sel_in_state value:     {} '
          '\naction_selection variable:  {} '
          '\naction_selection output:    {} '
          '\ncomparator sample:          {} '
          '\ncomparator target:          {} '
          '\nlearning mech act in:       {} '
          '\nlearning mech act out:      {} '
          '\nlearning mech error in:     {} '
          '\nlearning mech error out:    {} '
          '\nlearning mech learning_sig: {} '
          # '\npredicted reward:           {} '
        .format(
            action_selection.input_states[0].parameters.variable.get(context),
            action_selection.input_states[0].parameters.value.get(context),
            action_selection.parameters.variable.get(context),
            action_selection.output_state.parameters.value.get(context),
            comparator.input_states[pnl.SAMPLE].parameters.value.get(context),
            comparator.input_states[pnl.TARGET].parameters.value.get(context),
            learn_mech.input_states[pnl.ACTIVATION_INPUT].parameters.value.get(context),
            learn_mech.input_states[pnl.ACTIVATION_OUTPUT].parameters.value.get(context),
            learn_mech.input_states[pnl.ERROR_SIGNAL].parameters.value.get(context),
            learn_mech.output_states[pnl.ERROR_SIGNAL].parameters.value.get(context),
            learn_mech.output_states[pnl.LEARNING_SIGNAL].parameters.value.get(context),
            # action_selection.output_state.value[np.nonzero(action_selection.output_state.value)][0]
    ))


# Specify reward values associated with each action (corresponding to elements of esaction_selection.output_state.value)
# reward_values = [10, 0]
reward_values = [0, 10]

# Used by System to generate a reward on each trial based on the outcome of the action_selection (DDM) Mechanism
def reward(context=None):
    """Return the reward associated with the selected action"""
    selected_action = action_selection.output_state.parameters.value.get(context)
    if not any(selected_action):
        # Deal with initialization, during which action_selection.output_state.value may == [0,0]
        selected_action = np.array([1,0])
    return [reward_values[int(np.nonzero(selected_action)[0])]]


# Input stimuli for run of the System.
# Notes:
#    - for illustrative purposes, this list contains two sets of stimuli;
#        they will be used in sequence, and the sequence will be recycled for as many trials as specified by the
#        *num_trials* argument in the call the System's run method see below)
#    - rewards are specified by the reward function above
input_list = {input_layer: [[1, 1],[1, 1]]}

# # Shows graph of system (learning components are in orange)
# s.show_graph(show_learning=pnl.ALL, show_control=True, show_dimensions=True)
# s.show_graph(show_learning=pnl.ALL, show_mechanism_structure=True)
# s.show_graph(show_mechanism_structure=True)

# Run System.
# Note: *targets* is specified as the reward() function (see above).
# s.run(
#     num_trials=10,
#     inputs=input_list,
#     targets=reward,
#     call_before_trial=functools.partial(print_header, s),
#     call_after_trial=show_weights
# )

s.show_graph(show_learning=True)


from psyneulink.core.compositions.composition import Composition

__all__ = [
    'SystemComposition', 'SystemCompositionError'
]


class SystemCompositionError(Exception):

    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class SystemComposition(Composition):
    """

            Arguments
            ----------

            Attributes
            ----------

            Returns
            ----------
    """

    def __init__(self):
        super(SystemComposition, self).__init__()




import pytest

import doctest
import psyneulink as pnl


def test_mechanisms():
    fail, total = doctest.testmod(pnl.core.components.mechanisms.mechanism)
    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total),
                    pytrace=False)


def test_transfer_mechanism():
    fail, total = doctest.testmod(
            pnl.core.components.mechanisms.processing.transfermechanism)
    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total),
                    pytrace=False)


def test_integrator_mechanism():
    fail, total = doctest.testmod(
            pnl.core.components.mechanisms.processing.integratormechanism)
    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total),
                    pytrace=False)


# FAILS: AttributeError: 'InputState' object has no attribute '_name'
def test_objective_mechanism():
    fail, total = doctest.testmod(
            pnl.core.components.mechanisms.processing.objectivemechanism)
    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total),
                    pytrace=False)


def test_control_mechanism():
    fail, total = doctest.testmod(
            pnl.core.components.mechanisms.adaptive.control.controlmechanism)
    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total),
                    pytrace=False)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *********************************************  AdaptiveMechanism *****************************************************

"""

.. _AdaptiveMechanism_Overview:

Overview
--------

An AdaptiveMechanism is a type of `Mechanism <Mechanism>` that modifies the operation of one or more other `Components
<Component>`.  In general, an AdaptiveMechanism receives its input from an `ObjectiveMechanism`, however
this need not be the case.

.. _AdaptiveMechanism_Types:


There are four types of AdaptiveMechanism:

* `ModulatoryMechanism`
    takes an evaluative signal (generally received from an `ObjectiveMechanism`) and generates an
    `modulatory_allocation <ModulatoryMechanism.modulatory_allocation>`, each item of which is assigned to one of its
    `ModulatorySignals <ModulatorySignal>`;  each of those generates a `modulatory_signal
    <ModulatorySignal.modulatory_signal>` that is used by its `ModulatoryProjection(s) <ModulatoryProjection>` to
    modulate the parameter of a `function <State_Base.function>` (and thereby the `value <State_Base.value>`) of a
    `State`.  A ModulatoryMechanism can be assigned any combination of `ControlSignals <ControlSignal>` and
    `GatingSignals <GatingSignal>`.
..
* `ControlMechanism`
    a subclsass of `ModulatoryMechanism` that adds support for `costs <ControlMechanism.costs>`;  it takes an
    evaluative signal (generally received from an `ObjectiveMechanism`) and generates a `control_allocation
    <ControlMechanism.control_allocation>`, each item of which is assigned to one of its `ControlSignals
    <ControlSignal>`;  each of those generates a `control_signal <ControlSignal.control_signal>` that is used by its
    `ControlProjection(s) <ControlProjection>` to modulate the parameter of a `function <State_Base.function>` (and
    thereby the `value <State_Base.value>`) of a `State`.  A ControlMechanism can be assigned only the `ControlSignal`
    class of `ModulatorySignal`, but can be also be assigned other generic `OutputStates <OutputState>`.
..
* `GatingMechanism`
    a subclsass of `ModulatoryMechanism` that is specialized for modulating the input to or ouput from a `Mechanism`;
    it takes an evaluative signal (generally received from an `ObjectiveMechanism`) and generates a
    `gating_allocation <GatingMechanism.gating_allocation>`, each item of which is assigned to one of its
    `GatingSignals <ControlSignal>`;  each of those generates a `gating_signal <ControlSignal.control_signal>`
    that is used by its `GatingProjection(s) <ControlProjection>` to modulate the parameter of a `function
    <State_Base.function>` (and thereby the `value <State_Base.value>`) of an `InputState` or `OutputState`.
    A GatingMechanism can be assigned only the `GatingSignal` class of `ModulatorySignal`, but can be also be assigned
    other generic `OutputStates <OutputState>`.
.
..
* `LearningMechanism`
    takes an error signal (received from an `ObjectiveMechanism` or another `LearningMechanism`) and generates a
    `learning_signal <LearningMechanism.learning_signal>` that is provided to its `LearningSignal(s)
    <LearningSignal>`, and used by their `LearningProjections <LearningProjection>` to modulate the `matrix
    <MappingProjection.matrix>` parameter of a `MappingProjection`. A LearningMechanism can be assigned only
    `LearningSignals <LearningSignal>` as its `OuputStates <OutputState>`.

See `ModulatorySignal <ModulatorySignal_Naming>` for conventions used for the names of Modulatory components.

A single `AdaptiveMechanism` can be assigned more than one ModulatorySignal of the appropriate

which, each of which can
be assigned
different `allocations <ModulatorySignal.allocation>` (for ControlSignals and GatingSignals) or `learning_signals
<LearningMechanism.learning_signal>` (for LearningSignals).  A single ModulatorySignal can also be assigned multiple
ModulatoryProjections; however, as described  under `_ModulatorySignal_Projections`, they will all be assigned the
same `variable <ModulatoryProjection_Base.variable>`.


COMMENT:
AdaptiveMechanisms are always executed after all `ProcessingMechanisms <ProcessingMechanism>` in the `Process` or
`System` to which they belong have been executed, with all LearningMechanism executed first, then GatingMechanism,
ControlMechanism. All three types of AdaptiveMechanisms are executed before the next `TRIAL`, so that the
modifications they make are available during the next `TRIAL` run for the Process or System.
COMMENT

.. _AdaptiveMechanism_Creation:

Creating an AdaptiveMechanism
------------------------------

An AdaptiveMechanism can be created by using the standard Python method of calling the constructor for the desired type.
AdaptiveMechanisms of the appropriate subtype are also created automatically when other Components are created that
require them, or a form of modulation is specified for them. For example, a `ControlMechanism <ControlMechanism>` is
automatically created as part of a `System <System_Creation>` (for use as its `controller
<System.controller>`), or when `control is specified <ControlMechanism_Control_Signals>` for the parameter of a
`Mechanism <Mechanism>`; and one or more `LearningMechanism <LearningMechanism>` are created when learning is
specified for a `Process <Process_Learning_Sequence>` or a `System <System_Learning>` (see the documentation for
`subtypes <AdaptiveMechanism_Subtypes>` of AdaptiveMechanisms for more specific information about how to create them).

.. _AdaptiveMechanism_Structure:

Structure
---------

An AdaptiveMechanism has the same basic structure as a `Mechanism <Mechanisms>`.  In addition, every AdaptiveMechanism
has a `modulation <AdaptiveMechanism.modulation>` attribute, that determines the default method by which its
`ModulatorySignals <ModulatorySignal>` modify the value of the Components that they modulate (see the `modulation
<ModulatorySignal_Modulation>` for a description of how modulation operates, and the documentation for individual
subtypes of AdaptiveMechanism for more specific information about their structure and modulatory operation).

.. _AdaptiveMechanism_Execution:

Execution
---------

LearningMechanism and ControlMechanism are always executed at the end of a `TRIAL`, after all `ProcessingMechanisms
<ProcessingMechanism>` in the `Process` or `System` to which they belong have been executed; all LearningMechanism
executed first, and then ControlMechanism.  All modifications made are available during the next `TRIAL`.
GatingMechanism are executed in the same manner as ProcessingMechanisms;  however, because they almost invariably
introduce recurrent connections, care must be given to their `initialization and/or scheduling
<GatingMechanism_Execution>`).


.. _AdaptiveMechanism_Class_Reference:

Class Reference
---------------

"""

from psyneulink.core.components.mechanisms.mechanism import Mechanism_Base
from psyneulink.core.globals.keywords import ADAPTIVE_MECHANISM
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel

__all__ = [
    'AdaptiveMechanism_Base', 'AdaptiveMechanismError'
]


class AdaptiveMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value


class AdaptiveMechanism_Base(Mechanism_Base):
    """Subclass of `Mechanism <Mechanism>` that modulates the value(s) of one or more other `Component(s) <Component>`.

    .. note::
       AdaptiveMechanism is an abstract class and should NEVER be instantiated by a call to its constructor.
       They should be instantiated using the constructor for a `subclass <AdaptiveMechanism_Subtypes>`.

    COMMENT:

    Description:
        An AdaptiveMechanism is a Type of the `Mechanism <Mechanism>` Category of Component

    COMMENT


    Attributes
    ----------

    modulation : ModulationParam
        determines how the output of the AdaptiveMechanism's `ModulatorySignal(s) <ModulatorySignal>` are used to
        modulate the value of the State(s) to which their `ModulatoryProjection(s) <ModulatoryProjection>` project.
   """

    componentType = ADAPTIVE_MECHANISM

    class Parameters(Mechanism_Base.Parameters):
        """
            Attributes
            ----------

                modulation
                    see `modulation <AdaptiveMechanism_Base.modulation>`

                    :default value: None
                    :type:

        """
        modulation = None

    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TypeDefaultPreferences
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     kwPreferenceSetName: 'AdaptiveMechanismClassPreferences',
    #     kp<pref>: <setting>...}

    def __init__(self,
                 default_variable,
                 size,
                 modulation,
                 params,
                 name,
                 prefs,
                 context=None,
                 function=None,
                 **kwargs
                 ):
        """Abstract class for AdaptiveMechanism
        """

        if not hasattr(self, 'system'):
            self.system = None

        # Assign args to params and functionParams dicts
        params = self._assign_args_to_param_dicts(params=params,
                                                  modulation=modulation)

        super().__init__(default_variable=default_variable,
                         size=size,
                         params=params,
                         name=name,
                         prefs=prefs,
                         context=context,
                         function=function,
                         **kwargs
                         )



from . import composition
from . import pathwaycomposition
from . import systemcomposition
from . import compositionfunctionapproximator


from .composition import *
from .pathwaycomposition import *
from .systemcomposition import *
from .compositionfunctionapproximator import *

__all__ = list(composition.__all__)
__all__.extend(systemcomposition.__all__)
__all__.extend(pathwaycomposition.__all__)
__all__.extend(compositionfunctionapproximator.__all__)
